# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Contact: timhessels@hotmail.com
Repository: https://github.com/TimHessels/watertools
Module: Collect/CHIRPS

Description:
This module downloads daily and monthly CHIRPS 2.0 data from
ftp://chg-ftpout.geog.ucsb.edu server. Use the CHIRPS.daily or CHIRPS.monthly
functions to download and create daily or monthly CHIRPS images in Gtiff
format. The CHIRP data is available since 1981-01-01 till the present.

Examples:
from watertools.Collect import CHIRPS
CHIRPS.daily(Dir='C:/Temp/', Startdate='1999-02-01', Enddate='1999-02-03',
             latlim=[-10, 30], lonlim=[-20, 120])
CHIRPS.monthly(Dir='C:/Temp/', Startdate='1999-02-01', Enddate='1999-02-28',
               latlim=[-10, 30], lonlim=[-20, 120])
"""

from .daily import main as daily
from .monthly import main as monthly

__all__ = ['daily', 'monthly']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Functions

Description:
This module consists of the Functions that are used in the watertools
"""

from watertools.Functions import Area_Conversions, Random, Time_Conversions, Visualization

__all__ = ['Area_Conversions', 'Random', 'Time_Conversions', 'Visualization']

__version__ = '0.1'


# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Products/ETens

Description:
This module downloads ETensV1.0 data from
ftp.wateraccounting.unesco-ihe.org. Use the monthy function to
download and create monthly ET images in Gtiff format.
The data is available between 2003-01-01 till 2014-12-31.

Examples:
from watertools.Products import ETens
ETens.monthly(Dir='C:/Temp/', Startdate='2003-12-01', Enddate='2003-12-20',
           latlim=[41, 45], lonlim=[-8, -5])
"""

from .monthly import main as monthly

__all__ = ['monthly']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/MOD15

Description:
This module downloads MOD15 FPAR data from
http://e4ftl01.cr.usgs.gov/. Use the MOD15.FPAR_8daily function to
download and create 8 daily FPAR images in Gtiff format.
The data is available between 2000-02-18 till present.

Examples:
from watertools.Collect import MOD15
MOD15.FPAR_8daily(Dir='C:/Temp3/', Startdate='2003-12-01', Enddate='2003-12-20',
           latlim=[41, 45], lonlim=[-8, -5])
MOD15.LAI_8daily(Dir='C:/Temp3/', Startdate='2003-12-01', Enddate='2003-12-20',
           latlim=[41, 45], lonlim=[-8, -5])
"""

from .FPAR_8daily import main as FPAR_8daily
from .LAI_8daily import main as LAI_8daily

__all__ = ['FPAR_8daily', 'LAI_8daily']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Description:
This script collects ASCAT data from the VITO server. The data has a
daily temporal resolution and a spatial resolution of 0.25 degree. The
resulting tiff files are in the WGS84 projection.
The data is available between 2007-01-01 till present.

Example:
from watertools.Collect import ASCAT
ASCAT.daily(Dir='C:/Temp/', Startdate='2007-02-24', Enddate='2007-03-09',
                     latlim=[50,54], lonlim=[3,7])

"""
# General modules
import numpy as np
import os
import pandas as pd
import requests
from requests.auth import HTTPBasicAuth
import shutil
from netCDF4 import Dataset

# Water Accounting Modules
import watertools
import watertools.General.data_conversions as DC


def DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar):
    """
    This scripts downloads ASCAT SWI data from the VITO server.
    The output files display the Surface Water Index.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    lonlim -- [ymin, ymax]
    latlim -- [xmin, xmax]
    """

    # Check the latitude and longitude and otherwise reset lat and lon.
    if latlim[0] < -90 or latlim[1] > 90:
        print('Latitude above 90N or below 90S is not possible.\
            Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W.\
            Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)

    # Check Startdate and Enddate
    if not Startdate:
        Startdate = pd.Timestamp('2007-01-01')
    if not Enddate:
        Enddate = pd.Timestamp.now()

    # Make a panda timestamp of the date
    try:
        Enddate = pd.Timestamp(Enddate)
    except:
        Enddate = Enddate

    # amount of Dates weekly
    Dates = pd.date_range(Startdate, Enddate, freq='D')

    # Create Waitbar
    if Waitbar == 1:
        import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix='Progress:',
                                    suffix='Complete', length=50)

    # Define directory and create it if not exists
    output_folder = os.path.join(Dir, 'SWI', 'ASCAT', 'Daily')
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    output_folder_temp = os.path.join(Dir, 'SWI', 'ASCAT', 'Daily', 'Temp')
    if not os.path.exists(output_folder_temp):
        os.makedirs(output_folder_temp)

    # loop over dates
    for Date in Dates:

        # Define end filename
        End_filename = os.path.join(output_folder,
                                    'SWI_ASCAT_V3_Percentage_daily_%d.%02d.%02d.tif'
                                    % (Date.year, Date.month, Date.day))

        # Define IDs
        xID = 1800 + np.int16(np.array([np.ceil((lonlim[0])*10),
                                       np.floor((lonlim[1])*10)]))

        yID = np.int16(np.array([np.floor((-latlim[1])*10),
                                 np.ceil((-latlim[0])*10)])) + 900

        # Download the data from FTP server if the file not exists
        if not os.path.exists(End_filename):
            try:
                data = Download_ASCAT_from_VITO(End_filename,
                                                output_folder_temp, Date,
                                                yID, xID)
                # make geotiff file
                geo = [lonlim[0], 0.1, 0, latlim[1], 0, -0.1]
                DC.Save_as_tiff(name=End_filename, data=data,
                                geo=geo, projection="WGS84")
            except:
                print("Was not able to download file with date %s" % Date)

        # Adjust waitbar
        if Waitbar == 1:
            amount += 1
            WaitbarConsole.printWaitBar(amount, total_amount,
                                        prefix='Progress:', suffix='Complete',
                                        length=50)

    # remove the temporary folder
    shutil.rmtree(output_folder_temp)


def Download_ASCAT_from_VITO(End_filename, output_folder_temp, Date, yID, xID):
    """
    This function retrieves ALEXI data for a given date from the
    ftp.wateraccounting.unesco-ihe.org server.

    Restrictions:
    The data and this python file may not be distributed to others without
    permission of the WA+ team due data restriction of the ALEXI developers.

    Keyword arguments:

    """

    # Define date
    year_data = Date.year
    month_data = Date.month
    day_data = Date.day

    # filename of ASCAT data on server
    ASCAT_date = "%d%02d%02d1200" % (year_data, month_data, day_data)
    ASCAT_name = 'SWI_%s_GLOBE_ASCAT_V3.1.1' % ASCAT_date
    ASCAT_filename = "c_gls_SWI_%s_GLOBE_ASCAT_V3.1.1.nc" % ASCAT_date

    # Collect account and FTP information
    username, password = watertools.Functions.Random.Get_Username_PWD.GET("Copernicus")
    URL = "https://land.copernicus.vgt.vito.be/PDF/datapool/Vegetation/Soil_Water/SWI_V3/%s/%s/%s/%s/%s" % (year_data, month_data, day_data,
                                          ASCAT_name, ASCAT_filename)

    # Output zipfile
    output_ncfile_ASCAT = os.path.join(output_folder_temp, ASCAT_filename)

    # Download the ASCAT data
    try:
        y = requests.get(URL, auth=HTTPBasicAuth(username, password))
    except:
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        y = requests.get(URL, auth=(username, password), verify=False)

    # Write the file in system
    z = open(output_ncfile_ASCAT, 'wb')
    z.write(y.content)
    z.close()

    # Open nc file
    fh = Dataset(output_ncfile_ASCAT)
    dataset = fh.variables['SWI_010'][:,yID[0]:yID[1], xID[0]:xID[1]]
    data = np.squeeze(dataset.data, axis=0)
    data = data * 0.5
    data[data > 100.] = -9999
    fh.close()

    return(data)



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: watertools/Functions/Area_Conversions

"""


from watertools.Functions.Area_Conversions import Area_converter, Boundaries

__all__ = ['Area_converter', 'Boundaries']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
"""


from watertools.Functions.Time_Conversions import Eightdaily_to_monthly_state, Eightdaily_to_monthly_flux, Weekly_to_monthly_flux, Sixteendaily_to_monthly_state, Monthly_to_yearly_flux, Day_to_monthly_flux, Day_to_monthly_state, Monthly_to_yearly_state, Hour_to_daily_flux, Hour_to_daily_state

__all__ = ['Eightdaily_to_monthly_state', 'Eightdaily_to_monthly_flux', 'Weekly_to_monthly_flux', 'Sixteendaily_to_monthly_state', 'Monthly_to_yearly_flux', 'Day_to_monthly_flux', 'Day_to_monthly_state', 'Monthly_to_yearly_state', 'Hour_to_daily_flux', 'Hour_to_daily_state']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/JRC

Description:
This module downloads JRC water occurrence data from http://storage.googleapis.com/global-surface-water/downloads/.
Use the JRC.Occurrence function to
download and create a water occurrence image in Gtiff format.
The data represents the period 1984-2015.

Examples:
from watertools.Collect import JRC
JRC.Occurrence(Dir='C:/Temp3/', latlim=[41, 45], lonlim=[-8, -5])
"""

from .Occurrence import main as Occurrence

__all__ = ['Occurrence']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/ESACCI


Description:
This script automatically downloads ESACCI data 
"""

from .LU import main as LU

__all__ = ['LU']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/JRC
"""

# import general python modules
import os
import numpy as np
import shutil
import sys

# Water Accounting modules
import watertools.General.raster_conversions as RC
import watertools.General.data_conversions as DC

def DownloadData(Dir,latlim, lonlim, Waitbar):
    """
    This function downloads JRC data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    latlim -- [ymin, ymax] (values must be between -90 and 90)
    lonlim -- [xmin, xmax] (values must be between -180 and 180)
    Waitbar -- 1 (Default) will print a waitbar

    """

    # Check the latitude and longitude and otherwise set lat or lon on greatest extent
    if latlim[0] < -90 or latlim[1] > 90:
        print('Latitude above 90N or below 90S is not possible. Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W. Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)

    # Make directory for the JRC water occurrence data
    Dir = Dir.replace("/", os.sep)
    output_folder = os.path.join(Dir, 'JRC', 'Occurrence')
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    fileName_out = os.path.join(output_folder, 'JRC_Occurrence_percent.tif')

    if not os.path.exists(fileName_out):

        # Create Waitbar
        if Waitbar == 1:
            import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
            total_amount = 1
            amount = 0
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

        # This function defines the name of dataset that needs to be collected
        Names_to_download = Tiles_to_download(lonlim,latlim)

        # Pass variables to parallel function and run
        args = [output_folder, Names_to_download, lonlim, latlim]
        RetrieveData(args)

        if Waitbar == 1:
            amount = 1
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    else:
        print('JRC water occurrence map already exists')

    return()


def RetrieveData(args):
    """
    This function retrieves JRC data for a given date from the
    http://storage.googleapis.com/global-surface-water/downloads/ server.

    Keyword arguments:
    args -- A list of parameters defined in the DownloadData function.
    """
    # Argument
    [output_folder, Names_to_download, lonlim, latlim] = args

    # Collect the data from the JRC webpage and returns the data and lat and long in meters of those tiles
    try:
        Collect_data(Names_to_download, output_folder)
    except:
        print("Was not able to download the file")

    # Clip the data to the users extend
    if len(Names_to_download) == 1:
        trash_folder = os.path.join(output_folder, "Trash")
        data_in = os.path.join(trash_folder, Names_to_download[0])
        data_end, geo_end, Proj_out = RC.clip_data(data_in, latlim, lonlim)
    else:

        data_end = np.zeros([int((latlim[1] - latlim[0])/0.00025), int((lonlim[1] - lonlim[0])/0.00025)])

        for Name_to_merge in Names_to_download:
            trash_folder = os.path.join(output_folder, "Trash")
            data_in = os.path.join(trash_folder, Name_to_merge)
            geo_out, proj, size_X, size_Y = RC.Open_array_info(data_in)
            lat_min_merge = np.maximum(latlim[0], geo_out[3] + size_Y * geo_out[5])
            lat_max_merge = np.minimum(latlim[1], geo_out[3])
            lon_min_merge = np.maximum(lonlim[0], geo_out[0])
            lon_max_merge = np.minimum(lonlim[1], geo_out[0] + size_X * geo_out[1])

            lonmerge = [lon_min_merge, lon_max_merge]
            latmerge = [lat_min_merge, lat_max_merge]
            data_one, geo_one, proj_one = RC.clip_data(data_in, latmerge, lonmerge)

            Ystart = int((geo_one[3] - latlim[1])/geo_one[5])
            Yend = int(Ystart + np.shape(data_one)[0])
            Xstart = int((geo_one[0] - lonlim[0])/geo_one[1])
            Xend = int(Xstart + np.shape(data_one)[1])

            data_end[Ystart:Yend, Xstart:Xend] = data_one

        geo_end = tuple([lonlim[0], geo_one[1], 0, latlim[1], 0, geo_one[5]])

    # Save results as Gtiff
    fileName_out = os.path.join(output_folder, 'JRC_Occurrence_percent.tif')
    DC.Save_as_tiff(name=fileName_out, data=data_end, geo=geo_end, projection='WGS84')
    shutil.rmtree(trash_folder)
    return True



def Tiles_to_download(lonlim, latlim):
    '''
    Defines the JRC tiles that must be downloaded in order to cover the latitude and longitude limits

    Keywords arguments:
    lonlim -- [ymin, ymax] (longitude limits of the chunk or whole image)
    latlim -- [ymin, ymax] (latitude limits of the chunk or whole image)
    '''
    latmin = int(np.floor(latlim[0]/10.)*10)
    latmax = int(np.ceil(latlim[1]/10.)*10)
    lonmin = int(np.floor(lonlim[0]/10.)*10)
    lonmax = int(np.ceil(lonlim[1]/10.)*10)

    lat_steps = range(latmin + 10 , latmax + 10 , 10)
    lon_steps = range(lonmin, lonmax, 10)

    Names_to_download = []

    for lon_step in lon_steps:
        if lon_step < 0:
            string_long = "%sW" %abs(lon_step)
        else:
            string_long = "%sE" %lon_step
        for lat_step in lat_steps:
            if lat_step < 0:
                string_lat = "%sS" %abs(lat_step)
            else:
                string_lat = "%sN" %lat_step

            Name_to_download = "occurrence_%s_%sv1_3_2020.tif" %(string_long, string_lat)
            Names_to_download = np.append(Name_to_download, Names_to_download)

    return(Names_to_download)

def Collect_data(Names_to_download, output_folder):
    '''
    This function downloads all the needed JRC tiles from http://e4ftl01.cr.usgs.gov/MOLT/MOD13Q1.006/ as a hdf file.

    Keywords arguments:
    TilesHorizontal -- [TileMin,TileMax] max and min horizontal tile number
    TilesVertical -- [TileMin,TileMax] max and min vertical tile number
    Date -- 'yyyy-mm-dd'
    output_folder -- 'C:/file/to/path/'
    '''
    if sys.version_info[0] == 3:
        import urllib.parse
    if sys.version_info[0] == 2:
        import urllib

    for Name_to_download in Names_to_download:
        output_Trash = os.path.join(output_folder, "Trash")
        if not os.path.exists(output_Trash):
            os.mkdir(output_Trash)

        filename = os.path.join(output_Trash, Name_to_download)
        if not os.path.exists(filename):
           times = 0
           size = 0
           while times < 10 and size < 10000:
               url = "http://storage.googleapis.com/global-surface-water/downloads2020/occurrence/" + Name_to_download
               if sys.version_info[0] == 3:
                   code = urllib.request.urlopen(url).getcode()                   
               if sys.version_info[0] == 2:                   
                   code = urllib.urlopen(url).getcode()
               if (code != 404):
                  if sys.version_info[0] == 3:
                      urllib.request.urlretrieve(url, filename)
                  if sys.version_info[0] == 2:    
                      urllib.urlretrieve(url, filename)
                  times += 1
                  statinfo = os.stat(filename)
                  size = int(statinfo.st_size)
               else:
                  print(url + " not found")
                  times = 10

    return()



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/SoilGrids
"""
import os
from watertools.Collect.SoilGrids.DataAccess import DownloadData
import sys


def main(Dir, latlim, lonlim, level = 'sl1', Waitbar = 1):
    """
    Downloads SoilGrids data from ftp://ftp.soilgrids.org/data/recent/

    The following keyword arguments are needed:
    Dir -- 'C:/file/to/path/'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    level -- 'sl1' (Default) 
             'sl2'     
             'sl3'     
             'sl4'     
             'sl5'     
             'sl6'     
             'sl7'    
    Waitbar -- '1' if you want a waitbar (Default = 1)
    """

    # Create directory if not exists for the output
    output_folder = os.path.join(Dir, 'SoilGrids', 'Soil_Organic_Carbon_Content')
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Define the output map and create this if not exists
    nameEnd = os.path.join(output_folder, 'SoilOrganicCarbonContent_%s_SoilGrids_g_kg.tif' %level)    

    if not os.path.exists(nameEnd):

        # Create Waitbar
        if Waitbar == 1:
            print('\nDownload Soil Organic Carbon Content soil map of %s from SoilGrids.org' %level)
            import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
            total_amount = 1
            amount = 0
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

        # Download and process the data
        DownloadData(output_folder, latlim, lonlim, "SOC", level)

        if Waitbar == 1:
            amount = 1
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    else:
        if Waitbar == 1:
            print("\nSoil Organic Carbon Content soil map of %s from SoilGrids.org already exists in output folder" %level)
if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/MYD11

Description:
This module downloads MYD11 LST data from
http://e4ftl01.cr.usgs.gov/. Use the MYD11.LST_daily function to
download and create daily LST images in Gtiff format.
The data is available between 2000-02-18 till present.

Examples:
from watertools.Collect import MYD11
MYD11.LST_daily(Dir='C:/Temp3/', Startdate='2003-12-01', Enddate='2003-12-30',
           latlim=[41, 45], lonlim=[-8, -5])
"""

from .LST_daily import main as LST_daily
from .LST_8daily import main as LST_8daily

__all__ = ['LST_daily', 'LST_8daily']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/MYD9

Description:
This module downloads MYD9 reflectance data from
https://e4ftl01.cr.usgs.gov/. Use the MOD9.REF_daily function to
download and create daily Reflectance images in Gtiff format.
The data is available between 2000-02-18 till present.

Examples:
from watertools.Collect import MYD9
MYD9.REF_daily(Dir='C:/Temp3/', Startdate='2003-12-01', Enddate='2003-12-20',
           latlim=[41, 45], lonlim=[-8, -5])
"""

from .REF_daily import main as REF_daily

__all__ = ['REF_daily']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
WaterSat
author: Tim Martijn Hessels
Created on Mon Feb 18 20:50:36 2019
"""

import os
import urllib
from osgeo import gdal
import shutil
import numpy as np

def DownloadData(output_folder, latlim, lonlim):
    
    import watertools.General.data_conversions as DC 
    
    # Check the latitude and longitude and otherwise set lat or lon on greatest extent
    if latlim[0] < -90 or latlim[1] > 90:
        print('Latitude above 90N or below 90S is not possible. Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W. Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)  
        
    # Check output path and create if not exists
    output_folder_end = os.path.join(output_folder, "GlobCover", "Landuse")
    if not os.path.exists(output_folder_end):
        os.makedirs(output_folder_end)

    # Define end output
    filename_out_tiff = os.path.join(output_folder_end, "LC_GLOBCOVER_V2.3.tif")   

    if not os.path.exists(filename_out_tiff):

        # Define url where to download the globcover data
        url = r"http://due.esrin.esa.int/files/Globcover2009_V2.3_Global_.zip"
        
        # Create temp folder
        output_folder_temp = os.path.join(output_folder, "GlobCover", "Landuse", "Temp")
        if not os.path.exists(output_folder_temp):
            os.makedirs(output_folder_temp)
        
        # define output layer
        filename_out = os.path.join(output_folder_temp, "Globcover2009_V2.3_Global_.zip")
        
        # Download the data
        urllib.request.urlretrieve(url, filename=filename_out)
        
        # Extract data
        DC.Extract_Data(filename_out, output_folder_temp)
        
        # Define extracted tiff file
        globcover_filename = os.path.join(output_folder_temp, "GLOBCOVER_L4_200901_200912_V2.3.tif")
        
        # Open extract file
        dest = gdal.Open(globcover_filename)
        Array = dest.GetRasterBand(1).ReadAsArray()
        
        # Get information of geotransform and projection
        geo = dest.GetGeoTransform()
        proj = "WGS84"
        
        # define the spatial ids
        Xid = [np.floor((-geo[0] + lonlim[0])/geo[1]), np.ceil((-geo[0] + lonlim[1])/geo[1])]
        Yid = [np.floor((geo[3] - latlim[1])/-geo[5]), np.ceil((geo[3] - latlim[0])/-geo[5])]
        
        # Define the geotransform
        Xstart = geo[0] + Xid[0] * geo[1]
        Ystart = geo[3] + Yid[0] * geo[5]
        geo_out = tuple([Xstart, geo[1], 0, Ystart, 0, geo[5]])
        
        # Clip data out
        Array_end = Array[int(Yid[0]):int(Yid[1]), int(Xid[0]):int(Xid[1])]

        # Save data as tiff
        DC.Save_as_tiff(filename_out_tiff, Array_end, geo_out, proj)
        dest = None
        
        # remove temporary folder
        shutil.rmtree(output_folder_temp)
        
        


# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/Globcover
"""
from watertools.Collect.Globcover.DataAccess import DownloadData
import sys


def main(Dir, latlim, lonlim, Waitbar = 1):
    """
    Downloads Globcover data from http://due.esrin.esa.int/page_globcover.php

    The following keyword arguments are needed:
    Dir -- 'C:/file/to/path/'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    Waitbar -- '1' if you want a waitbar (Default = 1)
    """
    
    # Create Waitbar
    if Waitbar == 1:
        print('\nDownload Globcover landcover map')
        import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
        total_amount = 1
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    # Download and process the data
    DownloadData(Dir, latlim, lonlim)

    if Waitbar == 1:
        amount = 1
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
"""
Created on Wed Sep  9 08:28:08 2020

@author: timhe
"""
import os
import json

from cryptography.fernet import Fernet

def create_key():
    
    key_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), "wa_key.txt")
    
    if os.path.exists(key_filename):
        os.remove(key_filename)
    
    f = open(key_filename,"w+")
    f.write(str(Fernet.generate_key().decode("utf-8")))
    f.close()

    return()

def set_up_account(account="ALL"):
    
    key_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), "wa_key.txt")
    json_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), "keys.json")
    
    if not os.path.exists(key_filename):
        create_key()
    
    f = open(key_filename,"r")
    key = f.read()
    f.close()
    
    cipher_suite = Fernet(key.encode('utf-8'))

    if os.path.exists(json_filename):
    
        # open json file  
        with open(json_filename) as f:
            datastore = f.read()
        obj = json.loads(datastore)      
        f.close()
    else:
        obj = {}

    if account == "ALL":
        Servers = ["NASA", "NASA_BEARER", "GLEAM", "FTP_WA", "MSWEP", "Copernicus", "VITO", "WAPOR"]
    else:
        Servers = [account]
    
    for Server in Servers:
        
        if Server != "WAPOR" and Server != "NASA_BEARER":
            account_name = input("Type in your account username for %s" %Server)
            pwd = input("Type in your password for %s" %Server)        
            
            account_name_crypt = cipher_suite.encrypt(("%s" %account_name).encode('utf-8'))
            pwd_crypt = cipher_suite.encrypt(("%s" %pwd).encode('utf-8'))
            obj[Server] = ([str(account_name_crypt.decode("utf-8")), str(pwd_crypt.decode("utf-8"))])
        if  Server == "WAPOR" or Server == "NASA_BEARER":
            API_Key = input("Type in your API key for %s" %Server)
            API_Key_crypt = cipher_suite.encrypt(("%s" %API_Key).encode('utf-8'))
            obj[Server] = [str(API_Key_crypt.decode("utf-8"))]
            
    # save extent in task
    with open(json_filename, 'w') as outfile:
        json.dump(obj, outfile)        
    
    return()
    



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Contact: timhessels@hotmail.com
Repository: https://github.com/TimHessels/watertools
Module: Collect/GPM


Description:
This module downloads GPM (daily) and GPM (monthly) data from
ftp://disc2.nascom.nasa.gov. Use the GPM.daily or GPM.monthly functions to
download and create daily or monthly GPM images in Gtiff format.
The TRMM data is available since 2014-03-01 till present

Examples:
from watertools.Collect import GPM
GPM.daily(Dir='C:/Temp/', Startdate='2014-03-01', Enddate='2015-03-01',
           latlim=[-10, 30], lonlim=[-20, 120])
GPM.monthly(Dir='C:/Temp/', Startdate='2014-03-01', Enddate='2015-03-01',
             latlim=[-10, 30], lonlim=[-20, 120])
"""

from .daily import main as daily
from .monthly import main as monthly

__all__ = ['daily', 'monthly']

__version__ = '0.1'



# -*- coding: utf-8 -*-

# General modules
import numpy as np
import calendar
import os
import pandas as pd
import requests
import datetime
from joblib import Parallel, delayed
from netCDF4 import Dataset


# Water Accounting modules
import watertools.General.data_conversions as DC

'''
TimeCase = 'monthly'
Startdate = '2007-01-01'
Enddate = '2003-03-02'
Var = 'T'
Dir = 'C:\Test'
latlim = [30.1, 50.1]
lonlim = [-20.1, 20.1]
'''
def DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, Waitbar, cores,
                 TimeCase, CaseParameters):
    """
    This function downloads ECMWF six-hourly, daily or monthly data

    Keyword arguments:

    """

    # correct latitude and longitude limits
    latlim_corr_one = np.floor(latlim[0]/0.125) * 0.125
    latlim_corr_two = np.ceil(latlim[1]/0.125) * 0.125
    latlim_corr = [latlim_corr_one, latlim_corr_two]

    # correct latitude and longitude limits
    lonlim_corr_one = np.floor(lonlim[0]/0.125) * 0.125
    lonlim_corr_two = np.ceil(lonlim[1]/0.125) * 0.125
    lonlim_corr = [lonlim_corr_one, lonlim_corr_two]

    # Load factors / unit / type of variables / accounts
    VarInfo = VariablesInfo(TimeCase)
    Varname_dir = VarInfo.file_name[Var]

    # Create Out directory
    out_dir = os.path.join(Dir, "Weather_Data", "Model", "ECMWF", TimeCase, Varname_dir, "mean")
    if not os.path.exists(out_dir):
          os.makedirs(out_dir)

    DownloadType = VarInfo.DownloadType[Var]

    # Set required data for the three hourly option
    if TimeCase == 'six_hourly':
        string1 = 'oper'

    # Set required data for the daily option
    elif TimeCase == 'daily':
        Dates = pd.date_range(Startdate,  Enddate, freq='D')
    elif TimeCase == 'monthly':
        Dates = pd.date_range(Startdate,  Enddate, freq='MS')

    if DownloadType == 1:
        string1 = 'oper'
        string4 = "0"
        string6 = "00:00:00/06:00:00/12:00:00/18:00:00"
        string2 = 'sfc'
        string8 = 'an'

    if DownloadType == 2:
        string1 = 'oper'
        string4 = "12"
        string6 = "00:00:00/12:00:00"
        string2 = 'sfc'
        string8 = 'fc'

    if DownloadType == 3:
        string1 = 'oper'
        string4 = "0"
        string6 = "00:00:00/06:00:00/12:00:00/18:00:00"
        string2 = 'pl'
        string8 = 'an'

    string7 = '%s/to/%s'  %(Startdate, Enddate)

    parameter_number = VarInfo.number_para[Var]
    string3 = '%03d.128' %(parameter_number)
    string5 = '0.125/0.125'
    string9 = 'ei'
    string10 = '%s/%s/%s/%s' %(latlim_corr[1], lonlim_corr[0], latlim_corr[0], lonlim_corr[1])   #N, W, S, E


    # Download data by using the ECMWF API
    import watertools.Collect.ECMWF.ECMWFdownload as Download
    print('Use API ECMWF to collect the data, please wait')
    Download.API(Dir, DownloadType, string1, string2, string3, string4, string5, string6, string7, string8, string9, string10)

    # Open the downloaded data
    NC_filename = os.path.join(Dir,'data_interim.nc')
    fh = Dataset(NC_filename, mode='r')

    # Get the NC variable parameter
    parameter_var = VarInfo.var_name[Var]
    Var_unit = VarInfo.units[Var]
    factors_add = VarInfo.factors_add[Var]
    factors_mul = VarInfo.factors_mul[Var]

    # Open the NC data
    Data = fh.variables[parameter_var][:]
    Data_time = fh.variables['time'][:]
    lons = fh.variables['longitude'][:]
    lats = fh.variables['latitude'][:]

    # Define the georeference information
    Geo_four = np.nanmax(lats)
    Geo_one = np.nanmin(lons)
    Geo_out = tuple([Geo_one, 0.125, 0.0, Geo_four, 0.0, -0.125])

    # Create Waitbar
    if Waitbar == 1:
        import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    for date in Dates:

        # Define the year, month and day
        year =  date.year
        month =  date.month
        day =  date.day

        # Hours since 1900-01-01
        start = datetime.datetime(year=1900, month=1, day=1)
        end = datetime.datetime(year, month, day)
        diff = end - start
        hours_from_start_begin = diff.total_seconds()/60/60

        Date_good = np.zeros(len(Data_time))
        if TimeCase == 'daily':
             days_later = 1
        if TimeCase == 'monthly':
             days_later = calendar.monthrange(year,month)[1]

        Date_good[np.logical_and(Data_time>=hours_from_start_begin, Data_time<(hours_from_start_begin + 24 * days_later))] = 1

        Data_one = np.zeros([int(np.sum(Date_good)),int(np.size(Data,1)),int(np.size(Data,2))])
        Data_one = Data[np.int_(Date_good) == 1, :, :]

        # Calculate the average temperature in celcius degrees
        Data_end = factors_mul * np.nanmean(Data_one,0) + factors_add

        if VarInfo.types[Var] == 'flux':
            Data_end = Data_end * days_later

        VarOutputname = VarInfo.file_name[Var]

        # Define the out name
        name_out = os.path.join(out_dir, "%s_ECMWF_ERA-Interim_%s_%s_%d.%02d.%02d.tif" %(VarOutputname, Var_unit, TimeCase, year,month,day))

        # Create Tiff files
        DC.Save_as_tiff(name_out, Data_end, Geo_out, "WGS84")

        if Waitbar == 1:
            amount += 1
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)


    fh.close()

    return()

class VariablesInfo:
    """
    This class contains the information about the ECMWF variables
    http://rda.ucar.edu/cgi-bin/transform?xml=/metadata/ParameterTables/WMO_GRIB1.98-0.128.xml&view=gribdoc
    """
    number_para = {'T'  : 130,
                   '2T'  : 167,
                   'SRO' : 8,
                   'SSRO' :9,
                   'WIND' : 10,
                   '10SI' : 207,
                   'SP' : 134,
                   'Q' : 133,
                   'SSR': 176,
                   'R': 157,
                   'E':  182,
                   'SUND': 189,
                   'RO' : 205,
                   'TP' : 228,
                   '10U' : 165,
                   '10V' : 166,
                   '2D' : 168,
                   'SR' : 173,
                   'AL' : 174,
                   'HCC': 188}

    var_name	= {'T'  : 't',
                '2T'  : 't2m',
                'SRO' : 'sro',
                'SSRO' :'ssro',
                'WIND' : 'wind',
                '10SI' : '10si',
                'SP' :'sp',
                'Q' : 'q',
                'SSR': 'ssr',
                'R': 'r',
                'E':  'e',
                'SUND': 'sund',
                'RO' : 'ro',
                'TP' : 'tp',
                '10U' : 'u10',
                '10V' : 'v10',
                '2D' : 'd2m',
                'SR' : 'sr',
                'AL' : 'al',
                'HCC': 'hcc'}

	# ECMWF data
    descriptions = {'T'  : 'Temperature [K]',
                    '2T': '2 meter Temperature [K]',
                    'SRO' : 'Surface Runoff [m]',
                    'SSRO' :'Sub-surface Runoff [m]',
                    'WIND' : 'Wind speed [m s-1]',
                    '10SI' : '10 metre windspeed [m s-1]',
                    'SP' :'Surface Pressure [pa]',
                    'Q' : 'Specific humidity [kg kg-1]',
                    'SSR': 'Surface solar radiation [W m-2 s]',
                    'R': 'Relative humidity [%]',
                    'E':  'Evaporation [m of water]',
                    'SUND': 'Sunshine duration [s]',
                    'RO' : 'Runoff [m]',
                    'TP' : 'Total Precipitation [m]',
                    '10U' : '10 metre U wind component [m s-1]',
                    '10V' : '10 metre V wind component [m s-1]',
                    '2D' : '2 metre dewpoint temperature [K]',
                    'SR' : 'Surface roughness [m]',
                    'AL' : 'Albedo []',
                    'HCC': 'High cloud cover []'}

    # Factor add to get output
    factors_add = {'T': -273.15,
                   '2T': -273.15,
                   'SRO' : 0,
                   'SSRO' :0,
                   'WIND' : 0,
                   '10SI' : 0,
                   'SP' : 0,
                   'Q' : 0,
                   'SSR' : 0,
                   'R': 0,
                   'E': 0,
                   'SUND': 0,
                   'RO' : 0,
                   'TP' : 0,
                   '10U' : 0,
                   '10V' : 0,
                   '2D' : -273.15,
                   'SR' : 0,
                   'AL' : 0,
                   'HCC': 0}

    # Factor multiply to get output
    factors_mul = {'T': 1,
                   '2T': 1,
                   'SRO' : 1000,
                   'SSRO' :1000,
                   'WIND' : 1,
                   '10SI' : 1,
                   'SP' : 0.001,
                   'Q' : 1,
                   'SSR' : 1,
                   'R': 1,
                   'E': 1000,
                   'SUND': 1,
                   'RO' : 1000,
                   'TP' : 1000,
                   '10U' : 1,
                   '10V' : 1,
                   '2D' : 1,
                   'SR' : 1,
                   'AL' : 1,
                   'HCC': 1}

    types = {'T': 'state',
             '2T': 'state',
             'SRO' : 'flux',
             'SSRO' :'flux',
             'WIND' : 'state',
             '10SI' : 'state',
             'SP' : 'state',
             'Q' : 'state',
             'SSR' : 'state',
             'R': 'state',
             'E': 'flux',
             'SUND':'flux',
             'RO' : 'flux',
             'TP' : 'flux',
             '10U' : 'state',
             '10V' : 'state',
             '2D' : 'state',
             'SR' : 'state',
             'AL' : 'state',
             'HCC': 'state'}

    file_name = {'T': 'Tair2m',
                 '2T': 'Tair',
                 'SRO' : 'Surf_Runoff',
                 'SSRO' :'Subsurf_Runoff',
                 'WIND' : 'Wind',
                 '10SI' : 'Wind10m',
                 'SP' : 'Psurf',
                 'Q' : 'Qair',
                 'SSR' : 'SWnet',
                 'R': 'RelQair',
                 'E': 'Evaporation',
                 'SUND':'SunDur',
                 'RO' : 'Runoff',
                 'TP' : 'P',
                 '10U' : 'Wind_U',
                 '10V' : 'Wind_V',
                 '2D' : 'Dewpoint2m',
                 'SR' : 'SurfRoughness',
                 'AL' : 'Albedo',
                 'HCC': 'HighCloudCover'
                 }

    DownloadType = {'T': 3,
                 '2T': 1,
                 'SRO' : 0,
                 'SSRO' : 0,
                 'WIND' : 0,
                 '10SI' : 0,
                 'SP' : 1,
                 'Q' : 3,
                 'SSR' : 2,
                 'R': 3,
                 'E': 2,
                 'SUND':2,
                 'RO' : 2,
                 'TP' : 2,
                 '10U' : 1,
                 '10V' : 1,
                 '2D' : 1,
                 'SR' : 1,
                 'AL' : 1,
                 'HCC': 1
                 }

    def __init__(self, step):

        # output units after applying factor
        if step == 'six_hourly':
            self.units = {'T': 'C',
                          '2T': 'C',
                          'SRO' : 'mm',
                          'SSRO' :'mm',
                          'WIND' : 'm_s-1',
                          '10SI' : 'm_s-1',
                          'SP' : 'kpa',
                          'Q' : 'kg_kg-1',
                          'SSR' : 'W_m-2_s',
                          'R': 'percentage',
                          'E': 'mm',
                          'SUND':'s',
                          'RO' : 'mm',
                          'TP' : 'mm',
                          '10U' : 'm_s-1',
                          '10V' : 'm_s-1',
                          '2D' : 'C',
                          'SR' : 'm',
                          'AL' : '-',
                          'HCC': '-'
                 }

        elif step == 'daily':
            self.units = {'T': 'C',
                          '2T': 'C',
                          'SRO' : 'mm',
                          'SSRO' :'mm',
                          'WIND' : 'm_s-1',
                          '10SI' : 'm_s-1',
                          'SP' : 'kpa',
                          'Q' :'kg_kg-1',
                          'SSR' : 'W_m-2_s',
                          'R': 'percentage',
                          'E': 'mm',
                          'SUND':'s',
                          'RO' : 'mm',
                          'TP' : 'mm',
                          '10U' : 'm_s-1',
                          '10V' : 'm_s-1',
                          '2D' : 'C',
                          'SR' : 'm',
                          'AL' : '-',
                          'HCC': '-'}

        elif step == 'monthly':
            self.units = {'T': 'C',
                          '2T': 'C',
                          'SRO' : 'mm',
                          'SSRO' :'mm',
                          'WIND' : 'm_s-1',
                          '10SI' : 'm_s-1',
                          'SP' : 'kpa',
                          'Q' :'kg_kg-1',
                          'SSR' : 'W_m-2_s',
                          'R': 'percentage',
                          'E': 'mm',
                          'SUND':'s',
                          'RO' : 'mm',
                          'TP' : 'mm',
                          '10U' : 'm_s-1',
                          '10V' : 'm_s-1',
                          '2D' : 'C',
                          'SR' : 'm',
                          'AL' : '-',
                          'HCC': '-'}

        else:
            raise KeyError("The input time step is not supported")



# -*- coding: utf-8 -*-
"""
Created on Sun Dec 16 14:11:15 2018

@author: tih
"""
import numpy as np
import pandas as pd
import glob
import os
from osgeo import gdal

import watertools.General.raster_conversions as RC

def Visualize_Graph(input_folders, input_formats, Startdate, Enddate, input_shp, obj = 'id', Shp_value = 1, vmin = None, vmax = None):

    Dates = pd.date_range(Startdate, Enddate, freq = "D")
    
    Datasets_end = np.ones([len(Dates), len(input_folders)+1]) * np.nan
    
    for j in range(0,len(input_folders)):
    
        input_folder = input_folders[j]  
        input_format = input_formats[j]
        Dataset = Get_Dataset_Polygon(input_folder, input_format, Dates, input_shp, Shp_prop = obj, Shp_value = Shp_value)   
        
        if j == 0:
            
            Datasets_end[:,0:2] = Dataset
            
        else:
            Datasets_end[:,j + 1] = Dataset[:,1]
    
    import matplotlib.pyplot as plt
    
    colors = ['red','blue','green','black','yellow']
    
    for k in range(0,len(input_folders)):
        plt.scatter(Dates, Datasets_end[:,k+1], s =10, c = colors[k])
    plt.ylim(vmin, vmax)
    plt.legend(input_folders, loc='upper center',bbox_to_anchor=(0.5, -0.1))
    plt.show()  
    
    return(plt)
       
def Get_Dataset_Polygon(input_folder, input_format, Dates, input_shp, Shp_prop = 'id', Shp_value = 1, Stats = "mean"):
      
    Dataset = np.ones([len(Dates), 2]) * np.nan
    i = 0
    os.chdir(input_folder)
    
    filename_test = glob.glob('*.tif')[0]
    filename_in = os.path.join(input_folder, filename_test)
    dest = gdal.Open(filename_in) 
    geo = dest.GetGeoTransform()  
    
    epsg_tiff = RC.Get_epsg(dest)
    epsg_shp = RC.Get_epsg(input_shp, 'shp')
    
    if epsg_tiff != epsg_shp:
        input_shp = RC.reproject_shapefile(input_shp, epsg_tiff)
        
    # Create Mask of polygon
    mask_tiff = RC.GDAL_rasterize(input_shp, geo[1], Shp_prop)
    
    dest = RC.reproject_dataset_example(mask_tiff, filename_in)
           
    MASK = dest.GetRasterBand(1).ReadAsArray()
    MASK[MASK!=Shp_value] = np.nan           
    MASK[MASK==Shp_value] = 1

    for Date in Dates:
        print(Date)
        Day = Date.day
        Month = Date.month
        Year = Date.year
        DOY = Date.dayofyear
        
        filename = glob.glob(input_format.format(yyyy=Year, mm=Month, dd=Day, doy = DOY))
        
        Dataset[i,0] = Date.toordinal()
        
    
        if len(filename) > 0: 
            filename = filename[0]
           
            
            dest = gdal.Open(filename)
            Array = dest.GetRasterBand(1).ReadAsArray()
            Array = np.float_(Array)
            Array[Array==-9999] = np.nan
            try:
                if Stats == "mean":
                    Value = np.nanmean(Array[MASK==1])
                if Stats == "std":
                    Value = np.nanstd(Array[MASK==1])   
                if Stats == "max":
                    Value = np.nanmax(Array[MASK==1])                   
            except:
                dest = RC.reproject_dataset_example(mask_tiff, filename)
           
                MASK = dest.GetRasterBand(1).ReadAsArray()
                MASK[MASK>0] = 1
                MASK[MASK<=0] = np.nan   
                if Stats == "mean":
                    Value = np.nanmean(Array[MASK==1])
                if Stats == "std":
                    Value = np.nanstd(Array[MASK==1])                
                if Stats == "max":
                    Value = np.nanmax(Array[MASK==1])                  
                
            Dataset[i,1] = Value
        i += 1
        
    return(Dataset)


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 19 09:22:07 2017

@author: tih
"""

import numpy as np
import shapefile
from osgeo import gdal
import os

def Determine(Basin, Home_folder):
      			
    Shape_file_name_shp = os.path.join(Home_folder,'Basins', Basin + '.shp')
    if not os.path.exists(Shape_file_name_shp):
        print('%s is missing' %Shape_file_name_shp)
    Shape_file_name_dbf = os.path.join(Home_folder,'Basins', Basin + '.dbf')
    if not os.path.exists(Shape_file_name_dbf):
        print('%s is missing' %Shape_file_name_dbf)

    Basin_shp = shapefile.Reader(Shape_file_name_shp, Shape_file_name_dbf)
    Shape = Basin_shp.shapes()
    bbox = Shape[0].bbox
    Boundaries = dict()
    #Boundaries['Lonmin'] = np.floor(bbox[0]) - 0.1
    #Boundaries['Lonmax'] = np.ceil(bbox[2]) + 0.1
    #Boundaries['Latmin'] = np.floor(bbox[1]) - 0.1
    #Boundaries['Latmax'] = np.ceil(bbox[3]) + 0.1
    Boundaries['Lonmin'] = round(np.floor((bbox[0] * 10.) - 1.))/10.
    Boundaries['Lonmax'] = round(np.ceil((bbox[2] * 10.) + 1.))/10.
    Boundaries['Latmin'] = round(np.floor((bbox[1] * 10.) - 1.))/10.
    Boundaries['Latmax'] = round((np.ceil(bbox[3] * 10.) + 1.))/10.				
    return(Boundaries, Shape_file_name_shp)

def Determine_LU_Based(Basin, Home_folder):
    

    LU_file_name = os.path.join(Home_folder,'LU', Basin + '.tif')
    if not os.path.exists(LU_file_name):
        print('%s is missing' %LU_file_name)
        
    dest = gdal.Open(LU_file_name)
    Transform = dest.GetGeoTransform()
    sizeX = dest.RasterXSize
    sizeY = dest.RasterYSize
    
    #Boundaries['Lonmin'] = np.floor(bbox[0]) - 0.1
    #Boundaries['Lonmax'] = np.ceil(bbox[2]) + 0.1
    #Boundaries['Latmin'] = np.floor(bbox[1]) - 0.1
    #Boundaries['Latmax'] = np.ceil(bbox[3]) + 0.1
    Boundaries = dict()
    Boundaries['Lonmin'] = round(np.floor((Transform[0] * 10.) - 1.))/10.
    Boundaries['Lonmax'] = round(np.ceil(((Transform[0] + sizeX * Transform[1])  * 10.) + 1.))/10.
    Boundaries['Latmin'] = round(np.floor(((Transform[3] + sizeY * Transform[5])  * 10.) - 1.))/10.
    Boundaries['Latmax'] = round((np.ceil(Transform[3] * 10.) + 1.))/10.
				
    return(Boundaries, LU_file_name)


# -*- coding: utf-8 -*-
"""
Created on Thu Oct 22 16:14:43 2020

@author: timhe
"""
import numpy as np

def All_Arrays(y, x):
    
    return_dict = dict()
    paras = dir()
    for para in paras:
        if type(para) == str and para[0] != "_":
            print(para)
            A = eval(para)
            
            if type(A) == np.ndarray:
                if A.shape[0]>y and A.shape[1]>x:
                    return_dict[para] = A[y,x]
                    return_dict["Shape_%s"%para] = [A.shape]
                           
    return(return_dict)
                
                
        
        


# -*- coding: utf-8 -*-

# General modules
import numpy as np
import calendar
import os
import pandas as pd
import requests
from joblib import Parallel, delayed

# Water Accounting modules
import watertools
import watertools.General.data_conversions as DC

def DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, Waitbar, CaseParameters, cores, TimeCase):
    """
    This function downloads GLDAS CLSM daily data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Var -- 'wind_f_inst' : (string) For all variable codes: VariablesInfo('day').descriptions.keys()
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    cores -- 1....8
    CaseParameters -- See files: three_hourly.py, daily.py, and monthly.py
    """

    # Load factors / unit / type of variables / accounts
    VarInfo = VariablesInfo(TimeCase)
    username, password = watertools.Functions.Random.Get_Username_PWD.GET("NASA")

	# Set required data for the daily option
    # Set required data for the three hourly option
    if TimeCase == 'three_hourly':

        # Define output folder and create this one if not exists
        path = os.path.join(Dir, 'Weather_Data', 'Model', 'GLDAS',
                            TimeCase, Var)

        if not os.path.exists(path):
            os.makedirs(path)

        # Startdate if not defined
        sd_date = '1979-01-02'

        # Define Time frequency
        TimeFreq = 'D'

        # Define URL by using personal account
        #url = 'http://%s:%s@hydro1.gesdisc.eosdis.nasa.gov:80/dods/GLDAS_NOAH025SUBP_3H' %(username,password)
        url = 'https://hydro1.gesdisc.eosdis.nasa.gov/dods/GLDAS_CLM10SUBP_3H'  #%(username,password)

        # Name the definition that will be used to obtain the data
        RetrieveData_fcn = RetrieveData_three_hourly

        types = ['mean']

    elif TimeCase == 'daily':

        types = ['mean']

        # Define output folder and create this one if not exists
        path = {'mean': os.path.join(Dir, 'Weather_Data', 'Model', 'GLDAS_CLSM',
                                     TimeCase, Var, 'mean')}
        for i in range(len(types)):
            if not os.path.exists(path[types[i]]):
                os.makedirs(path[types[i]])

        # Startdate if not defined
        sd_date = '1948-01-01'

        # Define Time frequency
        TimeFreq = 'D'

        # Define URL by using personal account
        url = 'https://hydro1.gesdisc.eosdis.nasa.gov/dods/GLDAS_CLSM025_D.2.0'

        # Name the definition that will be used to obtain the data
        RetrieveData_fcn = RetrieveData_daily

    # Set required data for the monthly option
    elif TimeCase == 'monthly':

        types = ['mean']

        # Define output folder and create this one if not exists
        path = {'mean': os.path.join(Dir, 'Weather_Data', 'Model', 'GLDAS_CLSM',
                                     TimeCase, Var, 'mean')}
        for i in range(len(types)):
            if not os.path.exists(path[types[i]]):
                os.makedirs(path[types[i]])

        # Startdate if not defined
        sd_date = '1979-01-02'

        # Define Time frequency
        TimeFreq = 'MS'

        # Define URL by using personal account
        url = 'https://hydro1.gesdisc.eosdis.nasa.gov/dods/GLDAS_CLSM025_D.2.0'

        # Name the definition that will be used to obtain the data
        RetrieveData_fcn = RetrieveData_monthly
    # If none of the possible option are chosen
    else:
        raise KeyError("The input time interval is not supported")

    if TimeCase == 'three_hourly':

        # Define IDs (latitude/longitude)
        yID = np.int16(np.array([np.ceil((latlim[0] + 60)),
                                 np.floor((latlim[1] + 60))]))
        xID = np.int16(np.array([np.floor((lonlim[0] + 180)),
                                 np.ceil((lonlim[1] + 180))]))
    else:

        # Define IDs (latitude/longitude)
        yID = np.int16(np.array([np.ceil((latlim[0] + 60) * 4),
                                 np.floor((latlim[1] + 60) * 4)]))
        xID = np.int16(np.array([np.floor((lonlim[0] + 180) * 4),
                                 np.ceil((lonlim[1] + 180) * 4)]))
    
    # Check dates. If no dates are given, the max number of days is used.
    if not Startdate:
        Startdate = pd.Timestamp(sd_date)
    if not Enddate:
        Enddate = pd.Timestamp('Now')  # Should be much than available

    # Create all dates that will be calculated
    Dates = pd.date_range(Startdate, Enddate, freq=TimeFreq)

    # Create Waitbar
    if Waitbar == 1:
        import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    # Define the variable string name
    VarStr = VarInfo.names[Var]

    # Create one parameter with all the required arguments
    args = [path, url, Var, VarStr, VarInfo, TimeCase, xID, yID, lonlim, latlim, CaseParameters, username, password, types]

    # Pass variables to parallel function and run
    if not cores:
        for Date in Dates:
            RetrieveData_fcn(Date, args)
            if Waitbar == 1:
                amount += 1
                WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)
        results = True
    else:
        results = Parallel(n_jobs=cores)(delayed(RetrieveData_fcn)(Date, args)
                                         for Date in Dates)
    return results

def RetrieveData_three_hourly(Date, args):
    """
    This function retrieves GLDAS three-hourly data for a given date.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """
	
	# Open all the parameters
    [path, url, Var, VarStr, VarInfo, TimeCase, xID, yID, lonlim, latlim, CaseParameters, username, password,  types] = args

	# Open variable info parameters
    VarFactor = VarInfo.factors[Var]

	# Loop over the periods
    for period in CaseParameters:
        
        hour = (int(period)-1) * 3 
        
        # Check whether the file already exist or the worldfile is
        # downloaded
        BasinDir = path + '/' + VarStr + '_GLDAS-CLSM_' + \
            VarInfo.units[Var] + '_3hour_' + Date.strftime('%Y.%m.%d') + \
            '_%02d00.tif' %(hour)

        if not os.path.isfile(BasinDir):

            # Reset the begin parameters for downloading
            downloaded = 0
            N=0

            while downloaded == 0:
                try:

                    # Define time
                    zID = int(((Date - pd.Timestamp("1979-1-2")).days) * 8) + (period - 1)

                    # total URL
                    url_GLDAS = url + '.ascii?%s[%s][%s:1:%s][%s:1:%s]' %(Var,zID,yID[0],yID[1],xID[0],xID[1])

                    # open URL
                    try:
                        dataset = requests.get(url_GLDAS, allow_redirects=False, stream = True)
                    except:
                        from requests.packages.urllib3.exceptions import InsecureRequestWarning
                        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                        dataset = requests.get(url_GLDAS, allow_redirects=False, stream = True, verify = False)

                    try:
                        get_dataset = requests.get(dataset.headers['Location'], auth = (username,password), stream = True)
                    except:
                        from requests.packages.urllib3.exceptions import InsecureRequestWarning
                        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                        get_dataset = requests.get(dataset.headers['location'], auth = (username,password),stream = True, verify = False)

                                      # download data (first save as text file)
                    pathtext = os.path.join(path, 'temp%s.txt' % zID)
                    z = open(pathtext, 'w')
                    z.write(get_dataset.text)
                    z.close()

                    # Open text file and remove header and footer
                    data_start = np.genfromtxt(pathtext,dtype = float,skip_header = 1,skip_footer = 6,delimiter=',')
                    data = data_start[:,1:]

                    # Add the VarFactor
                    if VarFactor < 0:
                        data = data + VarFactor
                    else:
                        data = data * VarFactor
                    if VarInfo.types[Var] == 'flux':
                        data = data / 8

                    # Set Nan value for values lower than -9999
                    data[data < -9999] = -9999

                    # Say that download was succesfull
                    downloaded = 1

                # If download was not succesfull
                except:
                    data=[]

                    # Try another time
                    N = N + 1

                    # Stop trying after 10 times
                    if N == 10:
                        print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                        downloaded = 1

            # define geo
            lonlimGLDAS = xID[0] * 1.0 - 180
            latlimGLDAS = (yID[1] + 1) * 1.0 - 60

            # Save to geotiff file
            geo = [lonlimGLDAS,1.0,0,latlimGLDAS,0,-1.0]
            DC.Save_as_tiff(name=BasinDir, data=np.flipud(data[:,:]), geo=geo, projection="WGS84")

            # Delete data and text file
            del data
            os.remove(pathtext)

    return True
	
def RetrieveData_daily(Date, args):
    """
    This function retrieves GLDAS daily data for a given date.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """

    # Open all the parameters
    [path, url, Var, VarStr, VarInfo, TimeCase, xID, yID, lonlim, latlim, CaseParameters, username, password, types] = args

    # Reset the begin parameters for downloading
    downloaded = 0
    N = 0
    data_end = []

    # Check GLDAS version
    version = url[-3:]

    # Open all variable info
    for T in types:
        if T == 'mean':
            VarStr = VarInfo.names[Var]
        else:
            VarStr = VarInfo.names[Var] + '-' + T

        # Check whether the file already exist or
        # the worldfile is downloaded
        BasinDir = os.path.join(path[T], VarStr + '_GLDAS-CLSM_' +  VarInfo.units[Var] + '_daily_' + Date.strftime('%Y.%m.%d') + '.tif')

        # Check if the outputfile already excists
        if not os.path.isfile(BasinDir):

            # Create the time dimension
            if version == '2.0':
                zID_start = int(((Date - pd.Timestamp("1948-1-1")).days))
                zID_end = int(zID_start)
                if zID_end == 24472:
                    zID_end = 24470

            # define total url
            url_GLDAS = url + '.ascii?%s[%s:1:%s][%s:1:%s][%s:1:%s]' %(Var,zID_start,zID_end,yID[0],yID[1],xID[0],xID[1])

            # if not downloaded try to download file
            while downloaded == 0:
                try:

                    # open URL
                    try:
                        dataset = requests.get(url_GLDAS, allow_redirects=False,stream = True)
                    except:
                        from requests.packages.urllib3.exceptions import InsecureRequestWarning
                        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                        dataset = requests.get(url_GLDAS, allow_redirects=False,stream = True, verify = False)
                    try:
                        get_dataset = requests.get(dataset.headers['location'], auth = (username,password),stream = True)
                    except:
                        from requests.packages.urllib3.exceptions import InsecureRequestWarning
                        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                        get_dataset = requests.get(dataset.headers['location'], auth = (username,password),stream = True, verify = False)

                    # download data (first save as text file)
                    pathtext = os.path.join(path[T],'temp%s.txt' %str(zID_start))
                    z = open(pathtext,'w')
                    z.write(get_dataset.text)
                    z.close()

                    # Reshape data
                    datashape = [yID[1] - yID[0] + 1,xID[1] - xID[0] + 1]
                    data_start = np.genfromtxt(pathtext,dtype = float,skip_header = 1,skip_footer = 6,delimiter = ',')
                    data_list = np.asarray(data_start[:,1:])
                    data_end = np.resize(data_list,(datashape[0], datashape[1]))
                    os.remove(pathtext)

                    # Add the VarFactor
                    if VarInfo.factors[Var] < 0:
                        data_end[data_end != -9999] = data_end[data_end != -9999] + VarInfo.factors[Var]
                    else:
                        data_end[data_end != -9999] = data_end[data_end != -9999] * VarInfo.factors[Var]
                    data_end[data_end < -9999] = -9999

                    # define geo
                    lonlimGLDAS = xID[0] * 0.25 - 180
                    latlimGLDAS = (yID[1] + 1) * 0.25 - 60

                    # Download was succesfull
                    downloaded = 1

                # If download was not succesfull
                except:

                    # Try another time
                    N = N + 1

                    # Stop trying after 10 times
                    if N == 10:
                        print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                        downloaded = 1

            try:
                # Save to geotiff file

                if T == 'mean':
                    data = np.flipud(data_end)

                geo = [lonlimGLDAS,0.25,0,latlimGLDAS,0,-0.25]
                DC.Save_as_tiff(name=BasinDir, data=data, geo=geo, projection="WGS84")

            except:
                print('GLDAS map from '+ Date.strftime('%Y-%m-%d') + ' is not created')

    return True


def RetrieveData_monthly(Date, args):
    """
    This function retrieves GLDAS CLSM monthly data for a given date.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """

    # Open all the parameters
    [path, url, Var, VarStr, VarInfo, TimeCase, xID, yID, lonlim, latlim, CaseParameters, username, password, types] = args

    # Reset the begin parameters for downloading
    downloaded = 0
    N = 0
    data_end = []

    # Check GLDAS version
    version = url[-3:]

    # Open all variable info
    for T in types:
        if T == 'mean':
            VarStr = VarInfo.names[Var]
        else:
            VarStr = VarInfo.names[Var] + '-' + T

        # Check whether the file already exist or
        # the worldfile is downloaded
        BasinDir = os.path.join(path[T], VarStr + '_GLDAS-CLSM_' +  VarInfo.units[Var] + '_monthly_' + Date.strftime('%Y.%m.%d') + '.tif')

        # Check if the outputfile already excists
        if not os.path.isfile(BasinDir):

            # Create the time dimension
            if version == '2.0':
                zID_start = int(((Date - pd.Timestamp("1948-1-1")).days))
                Y = int(Date.year)
                M = int(Date.month)
                Mday = calendar.monthrange(Y,M)[1]
                zID_end = zID_start + Mday
                if zID_end == 24472:
                    zID_end = 24470
                    Mday = Mday - 2


            # define total url
            url_GLDAS = url + '.ascii?%s[%s:1:%s][%s:1:%s][%s:1:%s]' %(Var,zID_start,zID_end,yID[0],yID[1],xID[0],xID[1])

            # if not downloaded try to download file
            while downloaded == 0:
                try:

                    # open URL
                    try:
                        dataset = requests.get(url_GLDAS, allow_redirects=False,stream = True)
                    except:
                        from requests.packages.urllib3.exceptions import InsecureRequestWarning
                        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                        dataset = requests.get(url_GLDAS, allow_redirects=False,stream = True, verify = False)
                    try:
                        get_dataset = requests.get(dataset.headers['location'], auth = (username,password),stream = True)
                    except:
                        from requests.packages.urllib3.exceptions import InsecureRequestWarning
                        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                        get_dataset = requests.get(dataset.headers['location'], auth = (username,password),stream = True, verify = False)

                    # download data (first save as text file)
                    pathtext = os.path.join(path[T],'temp%s.txt' %str(zID_start))
                    z = open(pathtext,'w')
                    z.write(get_dataset.text)
                    z.close()

                    # Reshape data
                    datashape = [Mday,yID[1] - yID[0] + 1,xID[1] - xID[0] + 1]
                    data_start = np.genfromtxt(pathtext,dtype = float,skip_header = 1,skip_footer = 6,delimiter = ',')
                    data_list = np.asarray(data_start[:,1:])
                    data_end = np.resize(data_list,(Mday, datashape[1], datashape[2]))
                    os.remove(pathtext)

                    # Add the VarFactor
                    if VarInfo.factors[Var] < 0:
                        data_end[data_end != -9999] = data_end[data_end != -9999] + VarInfo.factors[Var]
                    else:
                        data_end[data_end != -9999] = data_end[data_end != -9999] * VarInfo.factors[Var]
                    data_end[data_end < -9999] = -9999

                    # define geo
                    lonlimGLDAS = xID[0] * 0.25 - 180
                    latlimGLDAS = (yID[1] + 1) * 0.25 - 60

                    # Download was succesfull
                    downloaded = 1

                # If download was not succesfull
                except:

                    # Try another time
                    N = N + 1

                    # Stop trying after 10 times
                    if N == 10:
                        print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                        downloaded = 1

            try:
                # Save to geotiff file
                if T == 'mean':
                    data_end[data_end<-100] = np.nan
                    data = np.flipud(np.nanmean(data_end, axis=0))
                if VarInfo.types[Var] == 'flux':
                    data = data * Mday

                geo = [lonlimGLDAS,0.25,0,latlimGLDAS,0,-0.25]
                DC.Save_as_tiff(name=BasinDir, data=data, geo=geo, projection="WGS84")

            except:
                print('GLDAS map from '+ Date.strftime('%Y-%m-%d') + ' is not created')

    return True

class VariablesInfo:
    """
    This class contains the information about the GLDAS variables
    """
    names = {'avgsurft_tavg': 'SurfaceTemperature',
             'canopint_tavg': 'TotCanopyWaterStorage',
             'evap_tavg': 'ET',
             'lwdown_f_tavg': 'LWdown',
             'lwnet_tavg': 'LWnet',
             'psurf_f_tavg': 'P',
             'qair_f_tavg': 'Hum',
             'qg_tavg': 'G',
             'qh_tavg': 'H',
             'qle_tavg': 'LE',
             'qs_tavg': 'Rsur',
             'qsb_tavg': 'Rsubsur',
             'qsm_tavg': 'SnowMelt',
             'rainf_f_tavg': 'P',
             'swe_tavg': 'SnowWaterEquivalent',
             'swdown_f_tavg': 'SWdown',
             'swnet_tavg': 'SWnet',
             'snowf_tavg': 'Snow',
             'soilmoist_s_tav': 'SoilMoisturSurface',
             'soilmoist_rz_ta': 'SoilMoistureRootZone',
             'soilmoist_p_tav': 'SoilMoistureProfile',
             'tair_f_tavg': 'Tair',
             'wind_f_tavg': 'W',
             'tveg_tavg' : 'Transpiration',
             'avgsurft': 'SurfaceTemperature',
             'canopint': 'TotCanopyWaterStorage',
             'evap': 'ET',
             'lwdown': 'LWdown',
             'lwnet': 'LWnet',
             'psurf': 'P',
             'qair': 'Hum',
             'qg': 'G',
             'qh': 'H',
             'qle': 'LE',
             'qs': 'Rsur',
             'qsb': 'Rsubsur',
             'qsm': 'SnowMelt',
             'rainf': 'P',
             'swe': 'SnowWaterEquivalent',
             'swdown': 'SWdown',
             'swnet': 'SWnet',
             'snowf': 'Snow',
             'tair': 'Tair',
             'wind': 'W',
             'gws_tavg': 'GWS'}
    descriptions = {'avgsurft_tavg': 'surface average surface temperature [k]',
                    'canopint_tavg': 'surface plant canopy surface water [kg/m^2]',
                    'evap_tavg': 'surface total evapotranspiration [kg/m^2/s]',
                    'lwdown_f_tavg': 'surface surface incident longwave radiation'
                               ' [w/m^2]',
                    'lwnet_tavg': 'surface net longwave radiation [w/m^2]',
                    'psurf_f_tavg': 'surface surface pressure [kPa]',
                    'qair_f_tavg': 'surface near surface specific humidity [kg/kg]',
                    'qg_tavg': 'surface ground heat flux [w/m^2]',
                    'qh_tavg': 'surface sensible heat flux [w/m^2]',
                    'qle_tavg': 'surface latent heat flux [w/m^2]',
                    'qs_tavg': 'storm surface runoff [kg/m^2/s]',
                    'qsb_tavg': 'baseflow-groundwater runoff [kg/m^2/s]',
                    'qsm_tavg': 'surface snowmelt [kg/m^2/s]',
                    'rainf_f_tavg': 'surface rainfall rate [kg/m^2/s]',
                    'swe_tavg': 'surface snow water equivalent [kg/m^2]',
                    'swdown_f_tavg': 'surface surface incident shortwave radiation'
                               ' [w/m^2]',
                    'swnet_tavg': 'surface net shortwave radiation [w/m^2]',
                    'snowf_tavg': 'surface snowfall rate [kg/m^2/s]',
                    'soilmoist_s_tav': 'surface soil moisture [kg/m^2]',
                    'soilmoist_rz_ta': 'root zone soil moisture [kg/m^2]',
                    'soilmoist_p_tav': 'profile soil moisture [kg/m^2]',
                    'tair_f_tavg': 'surface near surface air temperature [k]',
                    'wind_f_tavg': 'surface near surface wind speed [m/s]',
                    'tveg_tavg' : 'transpiration [w/m^2]',
                    'avgsurft': 'surface average surface temperature [k]',
                    'canopint': 'surface plant canopy surface water [kg/m^2]',
                    'evap': 'surface total evapotranspiration [kg/m^2/s]',
                    'lwdown': 'surface surface incident longwave radiation'
                               ' [w/m^2]',
                    'lwnet': 'surface net longwave radiation [w/m^2]',
                    'psurf': 'surface surface pressure [kPa]',
                    'qair': 'surface near surface specific humidity [kg/kg]',
                    'qg': 'surface ground heat flux [w/m^2]',
                    'qh': 'surface sensible heat flux [w/m^2]',
                    'qle': 'surface latent heat flux [w/m^2]',
                    'qs': 'storm surface runoff [kg/m^2/s]',
                    'qsb': 'baseflow-groundwater runoff [kg/m^2/s]',
                    'qsm': 'surface snowmelt [kg/m^2/s]',
                    'rainf': 'surface rainfall rate [kg/m^2/s]',
                    'swe': 'surface snow water equivalent [kg/m^2]',
                    'swdown': 'surface surface incident shortwave radiation'
                               ' [w/m^2]',
                    'swnet': 'surface net shortwave radiation [w/m^2]',
                    'snowf': 'surface snowfall rate [kg/m^2/s]',
                    'tair': 'surface near surface air temperature [k]',
                    'wind': 'surface near surface wind speed [m/s]',
                    'gws_tavg': 'ground water storage [mm]'}
    factors = {'avgsurft_tavg': -273.15,
               'canopint_tavg': 1,
               'evap_tavg': 86400,
               'lwdown_f_tavg': 1,
               'lwnet_tavg': 1,
               'psurf_f_tavg': 0.001,
               'qair_f_tavg': 1,
               'qg_tavg': 1,
               'qh_tavg': 1,
               'qle_tavg': 1,
               'qs_tavg': 86400,
               'qsb_tavg': 86400,
               'qsm_tavg': 86400,
               'rainf_f_tavg': 86400,
               'swe_tavg': 1,
               'swdown_f_tavg': 1,
               'swnet_tavg': 1,
               'snowf_tavg': 1,
               'soilmoist_s_tav': 1,
               'soilmoist_rz_ta': 1,
               'soilmoist_p_tav': 1,
               'tair_f_tavg': -273.15,
               'wind_f_tavg': 0.75,
               'tveg_tavg' : 1,
               'avgsurft': -273.15,
               'canopint': 1,
               'evap': 86400,
               'lwdown': 1,
               'lwnet': 1,
               'psurf': 0.001,
               'qair': 1,
               'qg': 1,
               'qh': 1,
               'qle': 1,
               'qs': 86400,
               'qsb': 86400,
               'qsm': 86400,
               'rainf': 86400,
               'swe': 1,
               'swdown': 1,
               'swnet': 1,
               'snowf': 1,
               'tair': -273.15,
               'wind': 1,
               'gws_tavg': 1}
    types = {'avgsurft_tavg': 'state',
             'canopint_tavg': 'state',
             'evap_tavg': 'flux',
             'lwdown_f_tavg': 'state',
             'lwnet_tavg': 'state',
             'psurf_f_tavg': 'state',
             'qair_f_tavg': 'state',
             'qg_tavg': 'state',
             'qh_tavg': 'state',
             'qle_tavg': 'state',
             'qs_tavg': 'flux',
             'qsb_tavg': 'flux',
             'qsm_tavg': 'flux',
             'rainf_f_tavg': 'flux',
             'swe_tavg': 'state',
             'swdown_f_tavg': 'state',
             'swnet_tavg': 'state',
             'snowf_tavg': 'state',
             'soilmoist_s_tav': 'state',
             'soilmoist_rz_ta': 'state',
             'soilmoist_p_tav': 'state',
             'tair_f_tavg': 'state',
             'wind_f_tavg': 'state',
             'tveg_tavg' : 'state',
             'avgsurft': 'state',
             'canopint': 'state',
             'evap': 'flux',
             'lwdown': 'state',
             'lwnet': 'state',
             'psurf': 'state',
             'qair': 'state',
             'qg': 'state',
             'qh': 'state',
             'qle': 'state',
             'qs': 'flux',
             'qsb': 'flux',
             'qsm': 'flux',
             'rainf': 'flux',
             'swe': 'state',
             'swdown': 'state',
             'swnet': 'state',
             'snowf': 'state',
             'tair': 'state',
             'wind': 'state',
             'gws_tavg': 'state'}

    def __init__(self, step):
        if step == 'three_hourly':
            self.units = {'avgsurft': 'C',
                          'canopint': 'mm',
                          'evap': 'mm-3H-1',
                          'lwdown': 'W-m-2',
                          'lwnet': 'W-m-2',
                          'psurf': 'kpa',
                          'qair': 'kg-kg',
                          'qg': 'W-m-2',
                          'qh': 'W-m-2',
                          'qle': 'W-m-2',
                          'qs': 'mm-3H-1',
                          'qsb': 'mm-3H-1',
                          'qsm': 'mm-3H-1',
                          'rainf': 'mm-3H-1',
                          'swe': 'mm',
                          'swdown': 'W-m-2',
                          'swnet': 'W-m-2',
                          'snowf': 'mm',
                          'tair': 'C',
                          'wind': 'm-s-1',
                          'gws_tavg': 'mm'}
        elif step == 'daily':
            self.units = {'avgsurft_tavg': 'C',
                          'canopint_tavg': 'mm',
                          'evap_tavg': 'mm-day-1',
                          'lwdown_f_tavg': 'W-m-2',
                          'lwnet_tavg': 'W-m-2',
                          'psurf_f_tavg': 'kpa',
                          'qair_f_tavg': 'kg-kg',
                          'qg_tavg': 'W-m-2',
                          'qh_tavg': 'W-m-2',
                          'qle_tavg': 'W-m-2',
                          'qs_tavg': 'mm-day-1',
                          'qsb_tavg': 'mm-day-1',
                          'qsm_tavg': 'mm-day-1',
                          'rainf_f_tavg': 'mm-day-1',
                          'swe_tavg': 'mm',
                          'swdown_f_tavg': 'W-m-2',
                          'swnet_tavg': 'W-m-2',
                          'snowf_tavg': 'mm',
                          'soilmoist_s_tav': 'kg-m-2',
                          'soilmoist_rz_ta': 'kg-m-2',
                          'soilmoist_p_tav': 'kg-m-2',
                          'tair_f_tavg': 'C',
                          'wind_f_tavg': 'm-s-1',
                          'tveg_tavg' : 'W-m-2',
                          'gws_tavg': 'mm'}
        elif step == 'monthly':
            self.units = {'avgsurft_tavg': 'C',
                          'canopint_tavg': 'mm',
                          'evap_tavg': 'mm-month-1',
                          'lwdown_f_tavg': 'W-m-2',
                          'lwnet_tavg': 'W-m-2',
                          'psurf_f_tavg': 'kpa',
                          'qair_f_tavg': 'kg-kg',
                          'qg_tavg': 'W-m-2',
                          'qh_tavg': 'W-m-2',
                          'qle_tavg': 'W-m-2',
                          'qs_tavg': 'mm-month-1',
                          'qsb_tavg': 'mm-month-1',
                          'qsm_tavg': 'mm-month-1',
                          'rainf_f_tavg': 'mm-month-1',
                          'swe_tavg': 'mm',
                          'swdown_f_tavg': 'W-m-2',
                          'swnet_tavg': 'W-m-2',
                          'snowf_tavg': 'mm',
                          'soilmoist_s_tav': 'kg-m-2',
                          'soilmoist_rz_ta': 'kg-m-2',
                          'soilmoist_p_tav': 'kg-m-2',
                          'tair_f_tavg': 'C',
                          'wind_f_tavg': 'm-s-1',
                          'tveg_tavg' : 'W-m-2',
                          'gws_tavg': 'mm'}
        else:
            raise KeyError("The input time step is not supported")



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Functions/Random/Start

"""
def printWaitBar(i, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█'):
    """
    This function will print a waitbar in the console

    Variables:

    i -- Iteration number
    total -- Total iterations
    fronttext -- Name in front of bar
    prefix -- Name after bar
    suffix -- Decimals of percentage
    length -- width of the waitbar
    fill -- bar fill
    """
    import sys

    # total can never be zero because we divide by total
    if total == 0:
        total = 0.0001

    percent = ("{0:." + str(decimals) + "f}").format(100 * (i / float(total)))
    filled = int(length * i // total)
    bar = fill * filled + '-' * (length - filled)

    sys.stdout.write('\r%s |%s| %s%% %s' %(prefix, bar, percent, suffix))
    sys.stdout.flush()

    if i == total:
        print()


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 28 07:54:17 2017

@author: tih
"""
import os
import sys
from watertools.Collect.ALEXI.DataAccess import DownloadData

def main(Dir, Startdate='', Enddate='', latlim=[-60, 70], lonlim=[-180, 180], Waitbar = 1):
    """
    This function downloads monthly ALEXI data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax] (values must be between -60 and 70)
    lonlim -- [xmin, xmax] (values must be between -180 and 180)
    """
    print('\nDownload monthly ALEXI evapotranspiration data for the period %s till %s' %(Startdate, Enddate))

    TimeStep = "weekly"
	
    # Download data
    DownloadData(Dir, Startdate, Enddate, latlim, lonlim, TimeStep, Waitbar)

    # Define directory
    Dir_ALEXI_Weekly = os.path.join(Dir, 'Evaporation', 'ALEXI', 'Weekly')
    Dir_ALEXI_Monthly = os.path.join(Dir, 'Evaporation', 'ALEXI', 'Monthly')

    # Create output directory
    if not os.path.exists(Dir_ALEXI_Monthly):
        os.mkdir(Dir_ALEXI_Monthly)

    # Create monthly maps based on weekly maps
    import watertools.Functions.Time_Conversions.Weekly_to_monthly_flux as Week2month
    Week2month.Nearest_Interpolate(Dir_ALEXI_Weekly, Startdate, Enddate, Dir_ALEXI_Monthly)

if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/MOD10
"""

# import general python modules
import os
import numpy as np
import pandas as pd
from osgeo import gdal
import urllib
from bs4 import BeautifulSoup
import re
import math
import datetime
import requests
import glob
from joblib import Parallel, delayed
import sys
if sys.version_info[0] == 3:
    import urllib.request    
    import urllib.parse
if sys.version_info[0] == 2:
    import urlparse

# Water Accounting modules
import watertools
import watertools.General.raster_conversions as RC
import watertools.General.data_conversions as DC

def DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar, cores, hdf_library, remove_hdf, period = "8-daily"):
    """
    This function downloads MOD10 8-daily data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax] (values must be between -90 and 90)
    lonlim -- [xmin, xmax] (values must be between -180 and 180)
    cores -- The number of cores used to run the routine. It can be 'False'
             to avoid using parallel computing routines.
	 nameDownload -- The name of the subset that must be download can be Fpar_500m or Lai_500m
    Waitbar -- 1 (Default) will print a waitbar
    """

    # Check start and end date and otherwise set the date to max
    if not Startdate:
        Startdate = pd.Timestamp('2000-02-18')
    if not Enddate:
        Enddate = pd.Timestamp('Now')


    # Make an array of the days of which the FPAR is taken
    if period == "8-daily":
        Dates = Make_TimeStamps(Startdate,Enddate)
    if period == "daily":
        Dates = pd.date_range(Startdate,Enddate,freq="D")
        
    # Create Waitbar
    if Waitbar == 1:
        import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    # Check the latitude and longitude and otherwise set lat or lon on greatest extent
    if latlim[0] < -90 or latlim[1] > 90:
        print('Latitude above 90N or below 90S is not possible. Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W. Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)

    # Make directory for the MODIS FPAR data
    Dir = Dir.replace("/", os.sep)
    if period == "8-daily":
        output_folder = os.path.join(Dir, 'MOD10')
    if period == "daily":    
        output_folder = os.path.join(Dir, 'MOD10', 'Daily')    

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # This function converts the values in the text file into horizontal and vertical number of the tiles which must be downloaded to cover the extent defined by the user
    TilesVertical, TilesHorizontal = watertools.Collect.MOD15.DataAccess.Get_tiles_from_txt(output_folder, hdf_library, latlim, lonlim)

    # Pass variables to parallel function and run
    args = [output_folder, TilesVertical, TilesHorizontal,lonlim, latlim, hdf_library, period]
    if not cores:
        for Date in Dates:
            RetrieveData(Date, args)
            if Waitbar == 1:
                amount += 1
                WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)
        results = True
    else:
        results = Parallel(n_jobs=cores)(delayed(RetrieveData)(Date, args)
                                         for Date in Dates)
    if remove_hdf == 1:
        # Remove all .hdf files
        os.chdir(output_folder)
        files = glob.glob("*.hdf")
        for f in files:
            os.remove(os.path.join(output_folder, f))

        # Remove all .txt files
        #files = glob.glob("*.txt")
        #for f in files:
        #   os.remove(os.path.join(output_folder, f))

    return(results)

def RetrieveData(Date, args):
    """
    This function retrieves MOD15 FPAR data for a given date from the
    http://e4ftl01.cr.usgs.gov/ server.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """
    # Argument
    [output_folder, TilesVertical, TilesHorizontal,lonlim, latlim, hdf_library, period] = args
    
    if period == "8-daily":
        FPARfileName = os.path.join(output_folder, 'SnowFrac_MOD10_-_8-daily_'  + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '.tif')
    if period == "daily":
        FPARfileName = os.path.join(output_folder, 'SnowFrac_MOD10_-_daily_'  + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '.tif')
 
    if not os.path.exists(FPARfileName):

        # Collect the data from the MODIS webpage and returns the data and lat and long in meters of those tiles
        try:
            Collect_data(TilesHorizontal, TilesVertical, Date, output_folder, hdf_library, period)
        except:
            print("Was not able to download the file")
        try:
            # Define the output name of the collect data function
            name_collect = os.path.join(output_folder, 'Merged.tif')
        
            # Reproject the MODIS product to epsg_to
            epsg_to ='4326'
            resolution = 0.005                
                    
            name_reprojected = RC.reproject_MODIS(name_collect, epsg_to, resolution = resolution)            

        
            # Clip the data to the users extend
            data, geo, proj = RC.clip_data(name_reprojected, latlim, lonlim)
        
            # Save the file as tiff
            DC.Save_as_tiff(name=FPARfileName, data=data, geo=geo, projection='WGS84')
        
            # remove the side products
            os.remove(os.path.join(output_folder, name_collect))
            os.remove(os.path.join(output_folder, name_reprojected))
                
        except:
            print("Failed for date: %s" %Date)
        
    return True

def Make_TimeStamps(Startdate,Enddate):
    '''
    This function determines all time steps of which the FPAR must be downloaded
    The time stamps are 8 daily.

    Keywords arguments:
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    '''

    # Define the DOY and year of the start day
    DOY = datetime.datetime.strptime(Startdate,'%Y-%m-%d').timetuple().tm_yday
    Year = datetime.datetime.strptime(Startdate,'%Y-%m-%d').timetuple().tm_year

    # Define the year of the end day
    YearEnd = datetime.datetime.strptime(Enddate,'%Y-%m-%d').timetuple().tm_year

    # Change the DOY of the start day into a DOY of MODIS day (8-daily) and create new startdate
    DOYstart = int(math.floor(DOY / 8.0) * 8) + 1
    DOYstart = str('%s-%s' %(DOYstart, Year))
    Day = datetime.datetime.strptime(DOYstart, '%j-%Y')
    Month = '%02d' % Day.month
    Day = '%02d' % Day.day
    Startdate = (str(Year) + '-' + str(Month) + '-' + str(Day))

    # Create the start and end data for the whole year
    YearStartDate = pd.date_range(Startdate, Enddate, freq = 'AS')
    YearEndDate = pd.date_range(Startdate, Enddate, freq = 'A')

    # Define the amount of years that are involved
    AmountOfYear = YearEnd - Year

    # If the startday is not in the same year as the enddate
    if AmountOfYear > 0:
        for i in range(0, AmountOfYear+1):
            if i == 0:
                Startdate1 = Startdate
                Enddate1 = YearEndDate[0]
                Dates = pd.date_range(Startdate1, Enddate1, freq = '8D')
            if i == AmountOfYear:
                Startdate1 = YearStartDate[-1]
                Enddate1 = Enddate
                Dates1 = pd.date_range(Startdate1, Enddate1, freq = '8D')
                Dates = Dates.union(Dates1)
            if (i != AmountOfYear and i != 0):
                Startdate1 = YearStartDate[i-AmountOfYear-1]
                Enddate1 = YearEndDate[i]
                Dates1 = pd.date_range(Startdate1, Enddate1, freq = '8D')
                Dates = Dates.union(Dates1)

    # If the startday is in the same year as the enddate
    if AmountOfYear == 0:
        Dates = pd.date_range(Startdate, Enddate, freq = '8D')

    return(Dates)


def Collect_data(TilesHorizontal,TilesVertical,Date,output_folder, hdf_library, period):
    '''
    This function downloads all the needed MODIS tiles from https://n5eil01u.ecs.nsidc.org/MOST/MOD10A2.006/ as a hdf file.

    Keywords arguments:
    TilesHorizontal -- [TileMin,TileMax] max and min horizontal tile number
    TilesVertical -- [TileMin,TileMax] max and min vertical tile number
    Date -- 'yyyy-mm-dd'
    output_folder -- 'C:/file/to/path/'
    '''

    # Make a new tile for the data
    sizeX = int((TilesHorizontal[1] - TilesHorizontal[0] + 1) * 2400)
    sizeY = int((TilesVertical[1] - TilesVertical[0] + 1) * 2400)
    DataTot = np.zeros((sizeY, sizeX))

    # Load accounts
    username, password = watertools.Functions.Random.Get_Username_PWD.GET('NASA')

    # Download the MODIS FPAR data
    if period == "8-daily":
        url = 'https://n5eil01u.ecs.nsidc.org/MOST/MOD10A2.006/' + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '/'
    if period == "daily":
        url = 'https://n5eil01u.ecs.nsidc.org/MOST/MOD10A1.061/' + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '/'
        
    dataset = requests.get(url, allow_redirects=False,stream = True)
    try:
        get_dataset = requests.get(dataset.headers['location'], auth = (username,password),stream = True).content
    except:
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        get_dataset  = requests.get(dataset.headers['location'], auth = (username, password), verify = False).content

    soup = BeautifulSoup(get_dataset, "lxml")

    if len(str(soup)) < 300:
        print('Download was not succesfull, please check NASA account')
        sys.exit(1)

    # Create the Lat and Long of the MODIS tile in meters
    for Vertical in range(int(TilesVertical[0]), int(TilesVertical[1])+1):
        Distance = 231.65635826395834*2 # resolution of a MODIS pixel in meter
        countY=(TilesVertical[1] - TilesVertical[0] + 1) - (Vertical - TilesVertical[0])

        for Horizontal in range(int(TilesHorizontal[0]), int(TilesHorizontal[1]) + 1):
            countX=Horizontal - TilesHorizontal[0] + 1

            for i in soup.findAll('a', attrs = {'href': re.compile('(?i)(hdf)$')}):

                # Find the file with the wanted tile number
                Vfile=str(i)[30:32]
                Hfile=str(i)[27:29]
                if int(Vfile) is int(Vertical) and int(Hfile) is int(Horizontal):

                    # Define the whole url name
                    if sys.version_info[0] == 3:
                        full_url = urllib.parse.urljoin(url, i['href'])

                    if sys.version_info[0] == 2:
                        full_url = urlparse.urljoin(url, i['href'])

		              # Reset the begin parameters for downloading
                    downloaded = 0
                    N=0

                    # if not downloaded try to download file
                    while downloaded == 0:

                        try:# open http and download whole .hdf
                            nameDownload_url = full_url
                            file_name = os.path.join(output_folder,nameDownload_url.split('/')[-1])
                            if os.path.isfile(file_name):
                                downloaded = 1
                            else:
                                x = requests.get(nameDownload_url, allow_redirects = False)
                                try:
                                    y = requests.get(x.headers['location'], auth = (username, password))
                                except:
                                    from requests.packages.urllib3.exceptions import InsecureRequestWarning
                                    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

                                    y = requests.get(x.headers['location'], auth = (username, password), verify = False)
                                z = open(file_name, 'wb')
                                z.write(y.content)
                                z.close()
                                statinfo = os.stat(file_name)
                                # Say that download was succesfull
                                if int(statinfo.st_size) > 1000:
                                     downloaded = 1

                        # If download was not succesfull
                        except:

                            # Try another time
                            N = N + 1

						      # Stop trying after 10 times
                        if N == 10:
                            print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                            downloaded = 1

                    try:
                        # Open .hdf only band with SnowFrac and collect all tiles to one array
                        scale_factor = 1
                        dataset = gdal.Open(file_name)
                        sdsdict = dataset.GetMetadata('SUBDATASETS')
                        sdslist = [sdsdict[k] for k in sdsdict.keys() if '_1_NAME' in k]
                        sds = []

                        for n in sdslist:
                            sds.append(gdal.Open(n))
                            full_layer = [i for i in sdslist if 'MOD_Grid_Snow_500m' in i]

                            idx = sdslist.index(full_layer[0])
                            if Horizontal == TilesHorizontal[0] and Vertical == TilesVertical[0]:
                                geo_t = sds[idx].GetGeoTransform()

                                # get the projection value
                                proj = sds[idx].GetProjection()

                            data = sds[idx].ReadAsArray()
                            countYdata = (TilesVertical[1] - TilesVertical[0] + 2) - countY
                            DataTot[int((countYdata - 1) * 2400):int(countYdata * 2400), int((countX - 1) * 2400):int(countX * 2400)]=data * scale_factor
                        del data

                    # if the tile not exists or cannot be opened, create a nan array with the right projection
                    except:
                        if Horizontal==TilesHorizontal[0] and Vertical==TilesVertical[0]:
                             x1 = (TilesHorizontal[0] - 19) * 2400 * Distance
                             x4 = (TilesVertical[0] - 9) * 2400 * -1 * Distance
                             geo = 	[x1, Distance, 0.0, x4, 0.0, -Distance]
                             geo_t=tuple(geo)

                        proj='PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",DATUM["Not specified (based on custom spheroid)",SPHEROID["Custom spheroid",6371007.181,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]'
                        data=np.ones((2400, 2400)) * (-9999)
                        countYdata=(TilesVertical[1] - TilesVertical[0] + 2) - countY
                        DataTot[(countYdata - 1) * 2400:countYdata * 2400,(countX - 1) * 2400:countX * 2400] = data * 0.01


    # Make geotiff file
    name2 = os.path.join(output_folder, 'Merged.tif')
    driver = gdal.GetDriverByName("GTiff")
    dst_ds = driver.Create(name2, DataTot.shape[1], DataTot.shape[0], 1, gdal.GDT_Float32, ['COMPRESS=LZW'])
    try:
        dst_ds.SetProjection(proj)
    except:
        proj='PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",DATUM["Not specified (based on custom spheroid)",SPHEROID["Custom spheroid",6371007.181,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]'
        x1 = (TilesHorizontal[0] - 18) * 2400 * Distance
        x4 = (TilesVertical[0] - 9) * 2400 * -1 * Distance
        geo = [x1, Distance, 0.0, x4, 0.0, -Distance]
        geo_t = tuple(geo)
        dst_ds.SetProjection(proj)

    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo_t)
    dst_ds.GetRasterBand(1).WriteArray(DataTot)
    dst_ds = None
    sds = None
    return()




# -*- coding: utf-8 -*-
import sys
from watertools.Collect.NLDAS.FORA_DataAccess import DownloadData


def main(Dir, Vars, Startdate, Enddate, latlim, lonlim, cores=False, Waitbar = 1, gldas_version = '2.1'):
    """
    This function downloads NLDAS monthly data for a given variable, time
    interval, and spatial extent.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Var --  ['tmp2m','spfh2m'] Variable code: VariablesInfo('day').descriptions.keys()
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    Waitbar -- 1 (Default) Will print a waitbar

    """
    for Var in Vars:

        if Waitbar == 1:
            print('\nDownloading monthly NLDAS %s data for the period %s till %s' %(Var, Startdate, Enddate))

        # Download data
        DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, Waitbar, cores,
                     TimeCase='monthly', CaseParameters=False)

if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/MOD16

Description:
This module downloads MOD16 ET data from
http://e4ftl01.cr.usgs.gov/. Use the MOD16.ET_monthly function to
download and create monthly ET images in Gtiff format.
The data is available between 2000-01-01 till 2014-12-31.

Examples:
from watertools.Collect import MOD16
MOD16.ET_monthly(Dir='C:/Temp3/', Startdate='2003-12-01', Enddate='2003-12-20',
           latlim=[41, 45], lonlim=[-8, -5])
"""

from .ET_monthly import main as ET_monthly
from .ET_8daily import main as ET_8daily

__all__ = ['ET_monthly', 'ET_8daily']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/ETmonitor

Restrictions:
The data and this python file may not be distributed to others without
permission of the WA+ team due data restriction of the ETmonitor developers.

Description:
This module downloads ETmonitor data from
ftp.wateraccounting.unesco-ihe.org. Use the ETmonitor.monthly function to
download and create monthly ETmonitor images in Gtiff format.
The data is available between 2008-01-01 till 2012-12-31.

Examples:
from watertools.Collect import ETmonitor
ETmonitor.ET_monthly(Dir='C:/Temp/', Startdate='2008-12-01', Enddate='2011-01-20',
           latlim=[-10, 30], lonlim=[-20, -10])
"""

from .ET_monthly import main as ET_monthly
from .ETpot_monthly import main as ETpot_monthly
from .Ei_monthly import main as Ei_monthly
from .Es_monthly import main as Es_monthly
from .Ew_monthly import main as Ew_monthly
from .Tr_monthly import main as Tr_monthly

__all__ = ['ET_monthly', 'ETpot_monthly','Ei_monthly', 'Es_monthly','Ew_monthly', 'Tr_monthly']

__version__ = '0.1'



# -*- coding: utf-8 -*-
'''
Authors: Tim Hessels
Module: Products/SoilGrids
'''

# import general python modules
import os
from osgeo import gdal
import numpy as np

# import watertools modules
from watertools.General import data_conversions as DC
from watertools.General import raster_conversions as RC

def Topsoil(Dir, latlim, lonlim, GF = True):
    """
    This function calculates the topsoil saturated soil characteristic (15cm)

    Keyword arguments:
    Dir -- 'C:/' path to the WA map
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    """
	
    print('/nCreate Theta Saturated map of the topsoil from SoilGrids')
    
    # Define parameters to define the topsoil
    SL = "sl3"
	
    Calc_Property(Dir, latlim, lonlim, SL, GF)
	
    return

def Subsoil(Dir, latlim, lonlim, GF = True):
    """
    This function calculates the subsoil saturated soil characteristic (100cm)

    Keyword arguments:
    Dir -- 'C:/' path to the WA map
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    """
	
    print('/nCreate Theta Saturated map of the subsoil from SoilGrids')
    
	# Define parameters to define the subsoil	
    SL = "sl6"
	
    Calc_Property(Dir, latlim, lonlim, SL, GF)
	
    return
	
def Calc_Property(Dir, latlim, lonlim, SL, GF = True):	
	
    import watertools.Collect.SoilGrids as SG

    # Download needed layers
    SG.Clay_Content(Dir, latlim, lonlim, level=SL)
    SG.Silt_Content(Dir, latlim, lonlim, level=SL)    
    SG.Organic_Carbon_Content(Dir, latlim, lonlim, level=SL)
    SG.Bulk_Density(Dir, latlim, lonlim, level=SL)
    
    # Define path to layers
    filename_clay = os.path.join(Dir, 'SoilGrids', 'Clay_Content' ,'ClayContentMassFraction_%s_SoilGrids_percentage.tif' %SL)
    filename_soc = os.path.join(Dir, 'SoilGrids', 'Soil_Organic_Carbon_Content' ,'SoilOrganicCarbonContent_%s_SoilGrids_g_kg.tif' %SL)
    filename_bulkdensity = os.path.join(Dir, 'SoilGrids', 'Bulk_Density' ,'BulkDensity_%s_SoilGrids_kg-m-3.tif' %SL)
    filename_silt = os.path.join(Dir, 'SoilGrids', 'Silt_Content' ,'SiltContentMassFraction_%s_SoilGrids_percentage.tif' %SL)
    
    # Define path for output
    if SL == "sl3":
        level = "Topsoil"
    elif SL == "sl6":
        level = "Subsoil"        

    filedir_out_densbulk = os.path.join( Dir, 'SoilGrids', 'Bulk_Density')
    if not os.path.exists(filedir_out_densbulk):
        os.makedirs(filedir_out_densbulk)
    filedir_out_thetasat = os.path.join(Dir, 'SoilGrids', 'Theta_Sat')
    if not os.path.exists(filedir_out_thetasat):
        os.makedirs(filedir_out_thetasat)    
    
    #filename_out_densbulk = os.path.join(filedir_out_densbulk ,'Bulk_Density_%s_SoilGrids_g-cm-3.tif' %level)
    filename_out_thetasat = os.path.join(filedir_out_thetasat,'Theta_Sat2_%s_SoilGrids_kg-kg.tif' %level)

    #if not (os.path.exists(filename_out_densbulk) and os.path.exists(filename_out_thetasat)):
    if not os.path.exists(filename_out_thetasat):
        
        # Open datasets
        dest_clay = gdal.Open(filename_clay)
        dest_soc = gdal.Open(filename_soc)
        dest_bulk = gdal.Open(filename_bulkdensity)
        dest_silt = gdal.Open(filename_silt)

        # Open Array info
        geo_out, proj, size_X, size_Y = RC.Open_array_info(filename_clay)    
        
        # Open Arrays
        Clay = dest_clay.GetRasterBand(1).ReadAsArray()
        SOC = dest_soc.GetRasterBand(1).ReadAsArray()
        Silt = dest_silt.GetRasterBand(1).ReadAsArray()        

        Clay = np.float_(Clay)
        '''
        try:
           Clay = RC.gap_filling(Clay, 0, method = 1)
        except:
            pass
        '''        
        OM = np.float_(SOC) * 1.72 # organic carbon to organic matter g/kg
        '''
        try:
           OM = RC.gap_filling(OM, 0, method = 1)
        except:
            pass
        '''
        Silt = np.float_(Silt)
        '''        
        try:
           Silt = RC.gap_filling(Silt, 0, method = 1)
        except:
            pass
        '''        
        Clay[Clay>100]=np.nan                
        Silt[Silt>100]=np.nan
        OM = OM/10  #g/kg to %
        
        Clay[Clay==0]=np.nan  
        Silt[Silt==0]=np.nan  
        OM[OM==0]=np.nan  
        
        # Calculate bulk density
        bulk_dens1 = dest_bulk.GetRasterBand(1).ReadAsArray()
        bulk_dens1 = bulk_dens1/1000 # kg/m2 to gr/cm3
        bulk_dens1 = np.float_(bulk_dens1)
        '''           
        try:
           bulk_dens1 = RC.gap_filling(bulk_dens1, 0, method = 1)
        except:
            pass
        '''                   
        bulk_dens2 = 1/(0.6117 + 0.3601 * Clay/100 + 0.002172 * np.power(Clay/100, 2)+ 0.01715 * np.log(OM))
    
        '''           
        try:
           bulk_dens2 = RC.gap_filling(bulk_dens2, 0, method = 1)
        except:
            pass       
        '''           
        bulk_dens = bulk_dens2 #np.where(bulk_dens1>bulk_dens2, bulk_dens1, bulk_dens2)
        
        '''
        # Oude methode gebaseerd op Schenost, Sinowski & Priesack (1996) 
        
        # Calculate theta saturated
        theta_sat = 0.85 * (1- (bulk_dens/2.65)) + 0.13 * Clay/100
        '''
        # Nieuwe methode gebaseerd op Toth et al (2014)
        
        # Calculate silt fraction based on clay fraction
        #Silt_fraction = 0.7 * (Clay/100) ** 2 + 0.308 * Clay/100
        Silt_fraction = Silt/100
        
        # Calculate theta sat
        theta_sat = 0.8308 - 0.28217 * bulk_dens + 0.02728 * Clay/100 + 0.0187 * Silt_fraction
        theta_sat[Clay==0] = -9999
        if GF == True:
            try:
               theta_sat = RC.gap_filling(theta_sat, -9999, method = 1)
            except:
                pass   

        # Save data
        #DC.Save_as_tiff(filename_out_densbulk, bulk_dens, geo_out, "WGS84")
        DC.Save_as_tiff(filename_out_thetasat, theta_sat, geo_out, "WGS84")
        
    return()    




# -*- coding: utf-8 -*-
"""
WaterSat
author: Tim Martijn Hessels
Created on Tue Feb 19 08:47:49 2019
"""

import os
import numpy as np
import pandas as pd
import datetime
import urllib
import requests
from netCDF4 import Dataset

import watertools

def DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, TimeStep, Period, Waitbar, data_type = ["mean"]):

	# watertools modules
    import watertools.General.data_conversions as DC
	
    # Check the latitude and longitude and otherwise set lat or lon on greatest extent
    if latlim[0] < -90 or latlim[1] > 90:
        print('Latitude above 90N or below 90S is not possible. Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W. Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)  
    
    # Get information of the parameter    
    VarInfo = VariablesInfo(TimeStep)
    Parameter = VarInfo.names[Var]
    unit  = VarInfo.units[Var]
    types  = VarInfo.types[Var]
    
    if TimeStep == "yearly":
        Parameter = "Temperature_Amplitude"
    
    # Create output folder
    output_folder = os.path.join(Dir, "MERRA", Parameter, TimeStep) 
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    
    if TimeStep.split("_")[-1] == "MERRA2":
        corrx = 0.625 * 0.5
        corry = 0.5 * 0.5
    else:
        corrx = 0
        corry = 0
    
    # Define IDs
    IDx = [np.floor((lonlim[0] + corrx + 180)/0.625), np.ceil((lonlim[1] + corrx + 180)/0.625)]
    IDy = [np.floor((latlim[0] + corry + 90)/0.5), np.ceil((latlim[1] + corry + 90)/0.5)]
    
    # Create output geo transform
    Xstart = -180 + 0.625 * IDx[0] - corrx
    Ystart = -90 + 0.5 * IDy[1] - corry
    
    geo_out = tuple([Xstart, 0.625, 0, Ystart, 0, -0.5])
    proj = "WGS84"
    
    if TimeStep == "yearly":
        Dates = pd.date_range(Startdate, Enddate, freq = "AS")
    else:
        Dates = pd.date_range(Startdate, Enddate, freq = "D")
        
    # Create Waitbar
    if Waitbar == 1:
        import watertools.Functions.Random.WaitbarConsole as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)
    
    for Date in Dates:
        
        # Define the IDz
        if TimeStep == "hourly_MERRA2":
            Hour = int((Period - 1) * 1)
            output_name = os.path.join(output_folder, "%s_MERRA_%s_hourly_%d.%02d.%02d_H%02d.M00.tif"%(Var, unit, Date.year, Date.month, Date.day, Hour))
            output_folder_temp = os.path.join(Dir, "MERRA", "Temp")
            if not os.path.exists(output_folder_temp):
                os.makedirs(output_folder_temp)
            year = Date.year
            month = Date.month
            day = Date.day
            output_name_min = output_folder
            output_name_max = output_folder
            
        if TimeStep == "daily_MERRA2":
            if "mean" in data_type:
                output_name = os.path.join(output_folder, "%s_MERRA_%s_daily_%d.%02d.%02d.tif" %(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name = output_folder
            if "min" in data_type:
                output_name_min = os.path.join(output_folder, "min", "%smin_MERRA_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name_min = output_folder
            if "max" in data_type:
                output_name_max = os.path.join(output_folder, "max", "%smax_MERRA_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name_max = output_folder
                
            output_folder_temp = os.path.join(Dir, "MERRA", "Temp")
            if not os.path.exists(output_folder_temp):
                os.makedirs(output_folder_temp)            
            year = Date.year
            month = Date.month
            day = Date.day
    
        if TimeStep == "three_hourly":
            IDz_start = IDz_end = int(((Date - pd.Timestamp("2002-07-01")).days) * 8) + (Period - 1)
            Hour = int((Period - 1) * 3)
            output_name = os.path.join(output_folder, "%s_MERRA_%s_3-hourly_%d.%02d.%02d_H%02d.M00.tif"%(Var, unit, Date.year, Date.month, Date.day, Hour))
            output_name_min = output_folder
            output_name_max = output_folder
            
        if TimeStep == "daily":
            IDz_start = int(((Date - pd.Timestamp("2002-07-01")).days) * 8) 
            IDz_end = IDz_start + 7
            if "mean" in data_type:
                output_name = os.path.join(output_folder, "%s_MERRA_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name = output_folder
            if "min" in data_type:
                output_name_min = os.path.join(output_folder, "min", "%smin_MERRA_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name_min = output_folder
            if "max" in data_type:
                output_name_max = os.path.join(output_folder, "max", "%smax_MERRA_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name_max = output_folder
                
        if TimeStep == "yearly":
            
            IDz_start = (Date.year - pd.Timestamp("2002-07-01").year) * 12 + Date.month - pd.Timestamp("2002-07-01").month 
            IDz_end = IDz_start + 11
            output_name = os.path.join(output_folder, "Tamp_MERRA_%s_yearly_%d.%02d.%02d.tif"%(unit, Date.year, Date.month, Date.day))
            output_name_min = output_folder
            output_name_max = output_folder
            
        if not (os.path.exists(output_name) and os.path.exists(output_name_min) and os.path.exists(output_name_max)):
            
            if (TimeStep == "hourly_MERRA2" or TimeStep == "daily_MERRA2"):
                
                if Date < datetime.datetime(1992,1,1):
                    number = 1
                elif (Date >= datetime.datetime(1992,1,1) and Date < datetime.datetime(2001,1,1)):
                    number = 2
                elif (Date >= datetime.datetime(2001,1,1) and Date < datetime.datetime(2011,1,1)):
                    number = 3
                else:
                    number = 4
                    
                if Date.month ==9 and Date.year==2020:
                    number2 = 1
                elif Date.month > 5 and Date.month < 10 and Date.year==2021:
                    number2 = 1                      
                else:
                    number2 = 0
                               
                if Var == "swgnet" or Var == "swgdn" or Var == "ts":
                    url_MERRA = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2T1NXRAD.5.12.4/%d/%02d/MERRA2_%s0%s.tavg1_2d_rad_Nx.%d%02d%02d.nc4" %(year, month, number, number2, year, month, day)
                elif Var == "prectotcorr":
                    url_MERRA = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2T1NXFLX.5.12.4/%d/%02d/MERRA2_%s0%s.tavg1_2d_flx_Nx.%d%02d%02d.nc4" %(year, month, number, number2, year, month, day)   
                else:    
                    url_MERRA = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2I1NXASM.5.12.4/%d/%02d/MERRA2_%s0%s.inst1_2d_asm_Nx.%d%02d%02d.nc4" %(year, month, number, number2, year, month, day)
            #https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2I1NXASM.5.12.4/2021/01/MERRA2_400.inst1_2d_asm_Nx.20210101.nc4
            #https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2I1NXASM.5.12.4/2021/09/MERRA2_401.inst1_2d_asm_Nx.20210902.nc4
            
            if (TimeStep == "three_hourly" or TimeStep == "daily"):                
                
                # define total url
                if (Var == "ps" or Var == "slp"):
                    url_start = r"https://opendap.nccs.nasa.gov/dods/GEOS-5/MERRAero/hourly/inst3hr_3d_asm_Nv."
                else:
                    url_start = r"https://opendap.nccs.nasa.gov/dods/GEOS-5/MERRAero/hourly/tavg3hr_2d_asm_Nx."
                
                url_MERRA = url_start + 'ascii?%s[%s:1:%s][%s:1:%s][%s:1:%s]' %(Var.replace("swgdn", "swgdwn"), IDz_start,IDz_end, int(IDy[0]),int(IDy[1]),int(IDx[0]),int(IDx[1]))
                print(url_MERRA)
            if TimeStep == "yearly":
                url_start = r"https://opendap.nccs.nasa.gov/dods/GEOS-5/MERRAero/monthly/tavg3hr_2d_asm_Nx."
                url_MERRA = url_start + 'ascii?%s[%s:1:%s][%s:1:%s][%s:1:%s]' %(Var, IDz_start,IDz_end, int(IDy[0]),int(IDy[1]),int(IDx[0]),int(IDx[1]))
                
                # Change date for now, there is no OpenDAP system for those years.
                if Date >= datetime.datetime(2015,1,1):
                    Date = datetime.datetime(2014,1,1)
                    
            # Reset the begin parameters for downloading
            downloaded = 0
            N = 0

            username, password = watertools.Functions.Random.Get_Username_PWD.GET("NASA")

            # if not downloaded try to download file
            while downloaded == 0:
                try:
                    N += 1
                    
                    if (TimeStep == "hourly_MERRA2" or TimeStep == "daily_MERRA2"):
                                          
                        # Define the output name that is downloaded
                        file_name = os.path.join(output_folder_temp, url_MERRA.split("/")[-1])
                        if not os.path.exists(file_name):
                        
                            # make contact with server
                            x = requests.get(url_MERRA, allow_redirects = False, timeout = 120)
                            try:
                                try:
                                    y = requests.get(x.headers['location'], auth = (username, password), timeout = 120)
                                except:
                                    from requests.packages.urllib3.exceptions import InsecureRequestWarning
                                    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
                                    y = requests.get(x.headers['location'], auth = (username, password), verify = False, timeout = 120)
    
                                
                                # Write the download in the output directory
                                z = open(file_name, 'wb')
                                z.write(y.content)
                                z.close()
                                statinfo = os.stat(file_name)
                                # Say that download was succesfull
                                if int(statinfo.st_size) > 1000:
                                     downloaded = 1  
                            except: 
                                
                                # Write the download in the output directory
                                z = open(file_name, 'wb')
                                z.write(x.content)
                                z.close()
                                statinfo = os.stat(file_name)
                                # Say that download was succesfull
                                if int(statinfo.st_size) > 1000:
                                     downloaded = 1                                                      
                                    
                        else:
                            downloaded = 1                             
                        
                        data_end, data_min, data_max = Get_NC_data_end(file_name, Var, TimeStep, Period, IDy, IDx, VarInfo)
                        #os.remove(file_name)
                                  
                    else:    
                            
                        # download data (first save as text file)
                        pathtext = os.path.join(output_folder,'temp%s.txt' %str(IDz_start))
                        
                        # Download the data
                        #urllib.request.urlretrieve(url_MERRA, filename=pathtext)       
        
                        if not os.path.exists(pathtext):
                        
                            # make contact with server
                            x = requests.get(url_MERRA, allow_redirects = False, timeout = 120)
            
                            # Write the download in the output directory
                            z = open(pathtext, 'wb')
                            z.write(x.content)
                            z.close()
                            statinfo = os.stat(pathtext)
                            # Say that download was succesfull
                            if int(statinfo.st_size) > 100:
                                 downloaded = 1                                                      
                                    
                        else:
                            downloaded = 1                                     
        
        
                        # Reshape data
                        datashape = [int(IDy[1] - IDy[0] + 1), int(IDx[1] - IDx[0] + 1)]
                        data_start = np.genfromtxt(pathtext,dtype = float,skip_header = 1,skip_footer = 6,delimiter = ',')
                        data_list = np.asarray(data_start[:,1:])
                        if TimeStep == "yearly":
                            data_end = np.resize(data_list,(12, datashape[0], datashape[1]))
                        if TimeStep == "daily":
                            data_end = np.resize(data_list,(8, datashape[0], datashape[1]))
                        if TimeStep == "three_hourly":
                            data_end = np.resize(data_list,(datashape[0], datashape[1]))
                        os.remove(pathtext)
        
                        # Set no data value
                        data_end[data_end>1000000] = -9999
                        
                        if TimeStep == "daily":
 
                            if "min" in data_type:
                                data_min = np.nanmin(data_end, 0)                            
                            
                            if "max" in data_type:
                                data_max = np.nanmax(data_end, 0)                            
                            
                            if "mean" in data_type:
                            
                                if types == "state":
                                    data_end = np.nanmean(data_end, 0)
                                else:
                                    data_end = np.nansum(data_end, 0)   
                                
                        if TimeStep == "yearly":
                            data_min = np.nanmin(data_end, 0)
                            data_max = np.nanmax(data_end, 0) 
                            data_end = data_max - data_min
        
                        # Download was succesfull
                        downloaded = 1

                    # Add the VarFactor
                    if VarInfo.factors[Var] < 0:
                        if "mean" in data_type:                       
                            data_end[data_end != -9999] = data_end[data_end != -9999] + VarInfo.factors[Var]
                            data_end[data_end < -9999] = -9999
                            data_end = np.flipud(data_end)
                        if "min" in data_type:
                            data_min[data_min != -9999] = data_min[data_min != -9999] + VarInfo.factors[Var]
                            data_min[data_min < -9999] = -9999
                            data_min = np.flipud(data_min)                                 
                        if "max" in data_type:
                            data_max[data_max != -9999] = data_max[data_max != -9999] + VarInfo.factors[Var]      
                            data_max[data_max < -9999] = -9999
                            data_max = np.flipud(data_max)                        
                                                    
                    else:
                        if "mean" in data_type: 
                            data_end[data_end != -9999] = data_end[data_end != -9999] * VarInfo.factors[Var]
                            data_end[data_end < -9999] = -9999
                            data_end = np.flipud(data_end)
                        if "min" in data_type:
                            data_min[data_min != -9999] = data_min[data_min != -9999] * VarInfo.factors[Var]
                            data_min[data_min < -9999] = -9999
                            data_min = np.flipud(data_min)                            
                        if "max" in data_type:
                            data_max[data_max != -9999] = data_max[data_max != -9999] * VarInfo.factors[Var]                          
                            data_max[data_max < -9999] = -9999
                            data_max = np.flipud(data_max)                        
                        
                    # Save as tiff file
                    if "mean" in data_type: 
                        DC.Save_as_tiff(output_name, data_end, geo_out, proj)
                    if "min" in data_type:
                        DC.Save_as_tiff(output_name_min, data_min, geo_out, proj)
                    if "max" in data_type:
                        DC.Save_as_tiff(output_name_max, data_max, geo_out, proj)
                    
                    # Stop trying after 3 times
                    if N == 4 and downloaded == 0:
                        print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                        downloaded = 1    
                    
                # If download was not succesfull
                except:
    
                    # Try another time
                    N = N + 1
    
                    # Stop trying after 3 times
                    if N == 4:
                        print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                        downloaded = 1
    
        if Waitbar == 1:
            amount += 1
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    return()    


def Get_NC_data_end(file_name,Var, TimeStep, Period, IDy, IDx, VarInfo):
                        
    dict_para = {'t2m': 'T2M',
         'u2m': 'U2M',
         'v2m': 'V2M',
         'q2m': 'QV2M',
         'tpw': 'TQV',
         'ps': 'PS',
         'slp': 'SLP',
         'swgnet': 'SWGDN',
         'swgdn': 'SWGDN',        
         'prectotcorr': 'PRECTOTCORR',
         'ts': 'TS'}  
      
    types  = VarInfo.types[Var]
    if TimeStep == "hourly_MERRA2":
        data_end = Dataset(file_name)["%s" %dict_para[Var]][int(Period-1),int(IDy[0]):int(IDy[1]),int(IDx[0]):int(IDx[1])]
        data_min = []
        data_max = []
        
    else:
        data = Dataset(file_name)["%s" %dict_para[Var]][:,int(IDy[0]):int(IDy[1]),int(IDx[0]):int(IDx[1])]
        if types == "state":
            data_end = np.nanmean(data, 0)
        else:
            data_end = np.nansum(data, 0)
            
        data[data==-9999] = np.nan
        data_min = np.nanmin(data, 0)
        data_max = np.nanmax(data, 0)     
        
    return(data_end, data_min, data_max)

class VariablesInfo:
    """
    This class contains the information about the GLDAS variables
    """
    names = {'t2m': 'Air_Temperature',
             'u2m': 'Eastward_Wind',
             'v2m': 'Northward_Wind',
             'q2m': 'Specific_Humidity',
             'tpw': 'Total_Precipitable_Water_Vapor',
             'ps': 'Surface_Pressure',
             'slp': 'Sea_Level_Pressure',
             'swgnet': 'Surface_Net_Downward_Shortwave_Flux',             
             'swgdn': 'Surface_Incoming_Shortwave_Flux',
             'prectotcorr': 'Total_Precipitation_Corrected',
             'ts': 'Surface_Skin_Temperature'
             }
    
    descriptions = {'t2m': '2m Air Temperature',
             'u2m': '2m Eastward wind',
             'v2m': '2m Northward wind',
             'q2m': '2m Specific Humidity',
             'tpw': 'Total Precipitable Water Vapor',
             'ps': 'Surface Pressure',
             'slp': 'Sea Level Pressure',
             'swgnet': 'Surface Net Downward Shortwave Flux',
             'swgdn': 'Surface Incoming Shortwave Flux',             
             'prectotcorr': 'Total Precipitation Corrected'  ,
             'ts': 'Surface Skin Temperature'
             }
    
    factors = {'t2m': 1,
             'u2m': 1,
             'v2m': 1,
             'q2m': 1,
             'tpw': 1,
             'ps': 0.001,
             'slp':  0.001,
             'swgnet':  1,
             'swgdn':  1,
             'prectotcorr':  3600,
             'ts': 1}
    
    types = {'t2m': 'state',
             'u2m': 'state',
             'v2m': 'state',
             'q2m': 'state',
             'tpw': 'state',
             'ps': 'state',
             'slp': 'state',
             'swgnet': 'state',
             'swgdn': 'state',
             'prectotcorr': 'flux',
             'ts':'state'
             }

    def __init__(self, step):
        if step == 'three_hourly':
            self.units = {'t2m': 'K',
             'u2m': 'm-s-1',
             'v2m': 'm-s-1',
             'q2m': 'kg-kg-1',
             'tpw': 'mm',
             'ps': 'kpa',
             'slp': 'kpa',
             'swgnet': 'W-m-2',
             'swgdn': 'W-m-2',             
             'prectotcorr': 'mm',
             'ts': 'K'
             }
            
        elif step == 'hourly_MERRA2':
            self.units = {'t2m': 'K',
             'u2m': 'm-s-1',
             'v2m': 'm-s-1',
             'q2m': 'kg-kg-1',
             'tpw': 'mm',
             'ps': 'kpa',
             'slp': 'kpa',
             'swgnet': 'W-m-2',
             'swgdn': 'W-m-2',
             'prectotcorr': 'mm',
             'ts': 'K'             
             }    
            
        elif (step == 'daily' or step == 'daily_MERRA2'):
            self.units = {'t2m': 'K',
             'u2m': 'm-s-1',
             'v2m': 'm-s-1',
             'q2m': 'kg-kg-1',
             'tpw': 'mm',
             'ps': 'kpa',
             'slp': 'kpa',
             'swgnet': 'W-m-2',
             'swgdn': 'W-m-2',             
             'prectotcorr': 'mm',
             'ts': 'K'             
             }
            
        elif step == 'yearly':
            self.units = {'t2m': 'K',
             'u2m': 'm-s-1',
             'v2m': 'm-s-1',
             'q2m': 'kg-kg-1',
             'tpw': 'mm',
             'ps': 'kpa',
             'slp': 'kpa',
             'swgnet': 'W-m-2',
             'swgdn': 'W-m-2',             
             'prectotcorr': 'mm',
             'ts': 'K'             
             }
        else:
            raise KeyError("The input time step is not supported")

